"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""

from collections import defaultdict
from typing import Dict, List

import numpy as np
from qrem.functions_qrem import povmtools, ancillary_functions as anf
from qrem.common import convert

import copy


# class KeyDependentDictForMarginals(defaultdict):
#     """
#     This is class used to store marginal probability distributions in dictionary.
#     It is "key dependent" in the sense that if user tries to refer to non-existing value for some
#     KEY, then this value is created as a marginal distribution which size depends on the KEY
#     NOTE: We assume that provided KEY is a string denoting  qubits subset
#     (see self.value_creating_function)


#     COPYRIGHT NOTE
#     The main idea of this code was taken from Reddit thread:
#     https://www.reddit.com/r/Python/comments/27crqg/making_defaultdict_create_defaults_that_are_a/

#     """

#     def __init__(self):
#         super().__init__(None)  # initialize as standard defaultdict

#         # This is the function which takes the string "key" that is assumed to label qubits subset
#         # in the form 'q2q3q11...' etc. It takes this key, calculates number of qubits N, and creates
#         # empty vector of the size d=2^N.
#         self.value_creating_function = lambda key: np.zeros(
#             (int(2 ** len(convert.get_qubit_indices_from_keystring(key))), 1),
#             dtype=float)

#     # called when key is missing
#     def __missing__(self, key):
#         # calculate the key-dependent value
#         ret = self.value_creating_function(key)
#         # put the value inside the dictionary
#         self[key] = ret
#         return ret


# class key_dependent_dict_for_marginals(defaultdict):
#     """
#     same as KeyDependentDictForMarginals but different name
#     TODO FBM: refactor this
#     """

#     def __init__(self):
#         super().__init__(None)  # initialize as standard defaultdict

#         # This is the function which takes the string "key" that is assumed to label qubits subset
#         # in the form 'q2q3q11...' etc. It takes this key, calculates number of qubits N, and creates
#         # empty vector of the size d=2^N.
#         self.value_creating_function = lambda key: np.zeros(
#             (int(2 ** len(convert.get_qubit_indices_from_keystring(key))), 1),
#             dtype=float)

#     # called when key is missing
#     def __missing__(self, key):
#         # calculate the key-dependent value
#         ret = self.value_creating_function(key)
#         # put the value inside the dictionary
#         self[key] = ret
#         return ret

def estimate_energy_from_marginals(weights_dictionary: Dict[str, float],
                                   marginals_dictionary: Dict[str, np.ndarray]):
    """
    Compute energy of Hamiltonian from dictionary of marginal distributions.

    :param weights_dictionary:
    :param marginals_dictionary:
    :return:
    """

    energy = 0
    for key_local_term in weights_dictionary.keys():
        weight = weights_dictionary[key_local_term]
        marginal = marginals_dictionary[key_local_term]

        qubits_number = int(np.log2(len(marginal)))

        for result_index in range(len(marginal)):
            bitstring = list(convert.integer_to_bitstring(result_index, qubits_number))
            bit_true = [int(x) for x in bitstring]
            parity = (-1) ** (np.count_nonzero(bit_true))
            energy += weight * marginal[result_index] * parity

    if isinstance(energy, list) or isinstance(energy, np.ndarray):
        return energy[0]
    else:
        return energy


import re


def get_energy_from_bitstring_diagonal(bitstring: str,
                                       weights_dict: Dict[str, float]
                                       ):
    energy = 0
    for qubits_string, hamiltonian_coefficient in weights_dict.items():
        qubit_indices = [int(q) for q in re.findall(r'\d+', qubits_string)]
        marginal_bitstring = [int(bitstring[q]) for q in qubit_indices]
        parity = (-1) ** (np.count_nonzero(marginal_bitstring))
        energy += parity * hamiltonian_coefficient

    return energy


def add_counts_from_dictionaries(counts_dictionaries_list: List[Dict[str, int]]) -> Dict[str, int]:
    """
    Merge multiple counts dictionaries.
    This is useful when you have results of multiple implementations of the same experiment.

    :param counts_dictionaries_list: list of results of counts dictionaries of the form:
                                    {'bitstring":number of measurements}
    :return:
    """

    # first dictionary will be template to which we will add counts
    merged_counts = copy.deepcopy(counts_dictionaries_list[0])

    # we go through all other dictionaries
    for counts_dictionary in counts_dictionaries_list[1:]:
        for bitstring, ticks in counts_dictionary.items():
            if bitstring in merged_counts.keys():
                merged_counts[bitstring] += ticks
            else:
                merged_counts[bitstring] = ticks

    return merged_counts


def convert_counts_overlapping_tomography(counts_dictionary: Dict[str, Dict[str, int]],
                                          experiment_name: str):
    def string_cutter(circuit_name: str):
        """
        This function cuts the name of the circuit to the format that will later be used by
        tomography data analyzers.


        :param circuit_name:
        It assumes the following convention:

        circuit_name = "experiment name" + "-" + "circuit label"+
        "no"+ "integer identifier for multiple implementations of the same circuit"

        for example the circuit can have name:
        "DDOT-010no3"

        which means that this experiment is Diagonal Detector Overlapping Tomography (DDOT),
        the circuit implements state "010" (i.e., gates iden, X, iden on qubits 0,1,2), and
        in the whole circuits sets this is the 4th (we start counting from 0) circuit that implements
        that particular state.

        :return:
        """
        experiment_string_len = len(list(experiment_name))
        full_name_now = circuit_name[experiment_string_len + 1:]
        new_string = ''
        for symbol_index in range(len(full_name_now)):
            if full_name_now[symbol_index] + full_name_now[symbol_index + 1] == 'no':
                break
            new_string += full_name_now[symbol_index]
        return new_string

    big_counts_dictionary = {}

    for circuit_name, counts_dict_now in counts_dictionary.items():
        proper_name_now = string_cutter(circuit_name)
        if proper_name_now not in big_counts_dictionary.keys():
            big_counts_dictionary[proper_name_now] = counts_dict_now
        else:
            big_counts_dictionary[proper_name_now] = add_counts_from_dictionaries(
                [big_counts_dictionary[proper_name_now], counts_dict_now])

    return big_counts_dictionary


import scipy as sc




"""qrem.providers.ibmutils.translate module contains helper functions for translating and preparation of circuits for running a characterisation/mitigation/benchmarking experiment
on IBM Quantum machines via QiskitRuntimeService or IBMServiceBackend.
"""

from typing import List, Union, Optional,Tuple
import numpy as np
from tqdm import tqdm
from pathlib import Path
# qiskit imports
from qiskit import transpile, QuantumCircuit, QuantumRegister, ClassicalRegister



from qrem.types import CircuitCollection
# qrem imports
from qrem.common.constants import EXPERIMENT_TYPE_SYMBLOS

def add_measurements(
        
    qiskit_circuit:QuantumCircuit,
    q_reg:QuantumRegister,
    c_reg:ClassicalRegister,
    valid_qubit_indices:List[int]):
    """
    Helper function that adds a measurement for every valid qubit defined in valid_qubit_indices list on existing qiskit_circuit (QuantumCircuit).
    """  
    # - we need to measure for every available qubit (number of qubits = classical_register_size = len(valid_qubit_indices)),
    # from a valid qubit with an index taken from a sorted list of valid_qubit_indices
    # - we only need as many qubits as max defined value of valid indices in our circuit definition (q_reg.size)

    for qubit_index in range(c_reg.size):
        which_qubit = q_reg[valid_qubit_indices[qubit_index]]
        which_classical_output = c_reg[qubit_index]
        qiskit_circuit.measure(which_qubit, which_classical_output)  

def circuits_translator_DDOT( circuit:np.ndarray, 
                                qiskit_circuit:QuantumCircuit,
                                valid_qubit_indices:list) -> QuantumCircuit:
    """
    Helper method that creates DDOT circuits in format native to QISKIT and applies to existing (empty) qiskit_circuit,
    based on the passed circuits and valid_qubit_indices, specifying what gates should be applied to what qubits.
    """

    
    for qubit_index in range(circuit.size):

        qubit_id = valid_qubit_indices[qubit_index]
        if circuit[qubit_index] == 0 :
            qiskit_circuit.id(qubit_id)
        elif circuit[qubit_index] == 1:
            qiskit_circuit.x(qubit_id)
        else:
            raise ValueError('Wrong circuit label: ', circuit)

    
    return qiskit_circuit

def circuits_translator_QDOT( circuit:np.ndarray, 
                                qiskit_circuit:QuantumCircuit,
                                valid_qubit_indices:list) -> QuantumCircuit:
    """
    Helper method that creates QDOT circuits in format native to QISKIT and applies to existing (empty) qiskit_circuit,
    based on the passed circuits and valid_qubit_indices, specifying what gates should be applied to what qubits.
    """
    # Opisac to lepiej i dodac ze to konkretnie dla Pauliego
        
    
    for qubit_index in range(circuit.size):
        qubit_id = valid_qubit_indices[qubit_index]

        if circuit[qubit_index] == 0:
            qiskit_circuit.id(qubit_id)
        elif circuit[qubit_index] == 1:
            qiskit_circuit.x(qubit_id)
        elif circuit[qubit_index] == 2:
            qiskit_circuit.h(qubit_id)
        elif circuit[qubit_index] == 3:
            qiskit_circuit.x(qubit_id)
            qiskit_circuit.h(qubit_id)
        elif circuit[qubit_index] == 4:
            qiskit_circuit.h(qubit_id)
            qiskit_circuit.s(qubit_id)
        elif circuit[qubit_index] == 5:
            qiskit_circuit.x(qubit_id)
            qiskit_circuit.h(qubit_id)
            qiskit_circuit.s(qubit_id)
        else:
            raise ValueError(
                f"Wrong circuit label: '{circuit[qubit_index]} from circuit: '{circuit}'")

        return qiskit_circuit




def transpile_circuits(qiskit_circuits:List[QuantumCircuit],
                       backend):
    """
    Helper method that transpiles circuits for the device. Optimization methodd set now to 0
    """    
    #TODO (PP) (MO) discuss optimization level here, should it be 0 or default 3
    return transpile(qiskit_circuits, 
                     backend = backend, 
                     optimization_level=0,
                     scheduling_method="asap")  



def download_results(provider, 
                    job_IDs:List[str],
                    backup: Union[bool,str,Path]=False,
                    overwrite:bool = False):    
    """
    Executes circuits prepared in QISKIT format (qiskit_circuits) on the backend provided by a qrem.common.providers.ibm.connection(...) function.

    Parameters
    ----------
    provider: IBMProvider
    job_IDs:List[str]

    backup: Union[bool,str,Path]=False
        Should original results be backed up somewhere (mostly for debugging)
    
    overwrite:bool
        Should original results overwrite any existing file

    Returns
    -------
    job_id_list: List[str]
        Lsit of submitted job IDs
    """         
    
    all_jobs_circuits = []
    all_jobs_results = []
    backup_dict={}

    #[1] Get results and circuits for every job ID
    for job_ID in job_IDs:
        job = provider.backend.retrieve_job(job_ID)
        result = job.result()

       
        #[1.1] retrive_circuits:
        circuits= job.circuits()
        job_circuits=[]
        for circuit in circuits:
            splitstr = circuit.name.split("___")
            label = splitstr[-2] #we set structure of  this label in translate_circuits_to_qiskit_format()
            id = splitstr[-1] #we set this label in translate_circuits_to_qiskit_format()
            job_circuits.append(label)
        all_jobs_circuits.append(job_circuits)
        
        
        #[1.1] retrieve counts:
        job_results = result.get_counts()
        all_jobs_results.append(job_results)

        if backup:
            backup_dict[job_ID] = result.to_dict()
            
    if backup:
        json_export_path = io.prepare_outfile(outpath = backup, overwrite = overwrite , default_filename= 'backup_ubm_original_results.pkl')
        
        json_data = orjson.dumps(backup_dict, default=lambda o: o.__dict__,
                    option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SORT_KEYS)
        
        with open(json_export_path, 'wb') as outfile:
            outfile.write(json_data)

    return all_jobs_results, all_jobs_circuits


def compare_circuits(circuit_labels :List[str],
                    original_circuits_path: Union[str,Path] = False,
                    original_circuits: Union[CircuitCollection,bool]= False):
    pass



# -*- coding: utf-8 -*-
"""
Created on Wed Nov  4 11:24:48 2015

@author: Peter Wittek
"""
from __future__ import division, print_function
from .sdp_relaxation import SdpRelaxation


class MoroderHierarchy(SdpRelaxation):

    """Class for obtaining a step in the Moroder hierarchy
    (`doi:10.1103/PhysRevLett.111.030501 <http://dx.doi.org/10.1103/PhysRevLett.111.030501>`_)
    :param variables: Commutative or noncommutative, Hermitian or nonhermiatian
                      variables, possibly a list of list of variables if the
                      hierarchy is not NPA.
    :type variables: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`
                     or a list of list.
    :param parameters: Optional symbolic variables for which moments are not
                       generated.
    :type parameters: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`
                     or a list of list.
    :param verbose: Optional parameter for level of verbosity:

                       * 0: quiet
                       * 1: verbose
                       * 2: debug level
    :type verbose: int.
    :param normalized: Optional parameter for changing the normalization of
                       states over which the optimization happens. Turn it off
                       if further processing is done on the SDP matrix before
                       solving it.
    :type normalized: bool.
    :param ppt: Optional parameter to impose a partial positivity constraint
                on the moment matrix.
    :type ppt: bool.

    Attributes:
      - `monomial_sets`: The monomial sets that generate the moment matrix
      blocks.

      - `monomial_index`: Dictionary that maps monomials to SDP variables.

      - `constraints`: The complete set of constraints after preprocesssing.

      - `primal`: The primal optimal value.

      - `dual`: The dual optimal value.

      - `x_mat`: The primal solution matrix.

      - `y_mat`: The dual solution matrix.

      - `solution_time`: The amount of time taken to solve the relaxation.

      - `status`: The solution status of the relaxation.

    """

    def __init__(self, variables, parameters=None, verbose=0, normalized=True,
                 ppt=False, parallel=False):
        super(MoroderHierarchy, self).__init__(variables, parameters, verbose,
                                               normalized, parallel)
        self.ppt = ppt

    def _generate_all_moment_matrix_blocks(self, n_vars, block_index):
        processed_entries = 0
        n_vars, block_index, _ = \
            self._generate_moment_matrix(n_vars, block_index,
                                         processed_entries,
                                         self.monomial_sets[0],
                                         self.monomial_sets[1],
                                         ppt=self.ppt)
        return n_vars, block_index

    def _calculate_block_structure(self, inequalities, equalities,
                                   momentinequalities, momentequalities,
                                   extramomentmatrix, removeequalities,
                                   block_struct=None):
        """Calculates the block_struct array for the output file.
        """
        block_struct = []
        if self.verbose > 0:
            print("Calculating block structure...")
        block_struct.append(len(self.monomial_sets[0]) *
                            len(self.monomial_sets[1]))
        if extramomentmatrix is not None:
            for _ in extramomentmatrix:
                block_struct.append(len(self.monomial_sets[0]) *
                                    len(self.monomial_sets[1]))
        super(MoroderHierarchy, self).\
            _calculate_block_structure(inequalities, equalities,
                                       momentinequalities, momentequalities,
                                       extramomentmatrix,
                                       removeequalities,
                                       block_struct=block_struct)

    def _estimate_n_vars(self):
        self.n_vars = 0
        if self.parameters is not None:
            self.n_vars = len(self.parameters)
        n_monomials = len(self.monomial_sets[0])*len(self.monomial_sets[1])
        self.n_vars += int(n_monomials * (n_monomials + 1) / 2)



# -*- coding: utf-8 -*-
"""
The module contains helper functions to calculate the chordal extension of the
correlative sparsity pattern matrix. It is largely based on the MATLAB version
in SparsePOP.

Created on Sun Nov 30 15:01:13 2014

@author: Peter Wittek
"""
from __future__ import division, print_function
import numpy as np
import random
from .nc_utils import get_support, flatten


def sliding_cliques(k, n):
    one_sides = []
    cliques = []
    for i in range(n-int(k/2)+1):
        side = [0] * n
        for j in range(int(k/2)):
            side[i+j] = 1
        one_sides.append(side)
    for side1 in one_sides:
        for side2 in one_sides:
            clique = list(side1)
            clique.extend(side2)
            cliques.append(clique)
    return cliques


def _generate_clique(variables, obj, inequalities, equalities,
                     momentinequalities, momentequalities):
    n_dim = len(variables)
    rmat = np.eye(n_dim)
    # Objective: if x_i & x_j in monomial, rmat_ij = rand
    if obj is not None:
        for support in get_support(variables, obj):
            nonzeros = np.nonzero(support)[0]
            value = random.random()
            for i in nonzeros:
                for j in nonzeros:
                    rmat[i, j] = value
    # Constraints: if x_i & x_j in support, rmat_ij = rand
    for polynomial in flatten([inequalities, equalities, momentinequalities,
                               momentequalities]):
        support = np.any(get_support(variables, polynomial), axis=0)
        nonzeros = np.nonzero(support)[0]
        value = random.random()
        for i in nonzeros:
            for j in nonzeros:
                rmat[i, j] = value
    rmat = rmat + 5*n_dim*np.eye(n_dim)
    # TODO: approximate minimum degree ordering should go before the Cholesky
    # decomposition
    R = np.linalg.cholesky(rmat).T
    R[np.nonzero(R)] = 1
    remaining_indices = [0]
    for i in range(1, n_dim):
        check_set = R[i, i:n_dim]
        one = np.nonzero(check_set)[0]
        n_ones = len(one)
        clique_result = np.dot(R[:i, i:n_dim], check_set.T)
        test = False
        for t in clique_result:
            if t == n_ones:
                test = True
                break
        if not test:
            remaining_indices.append(i)
    clique_set = R[remaining_indices, :]
    return clique_set


def _generate_clique_alt(variables, obj, inequalities, equalities,
                         momentinequalities, momentequalities):
    n_dim = len(variables)
    rmat = spmatrix(1.0, range(n_dim), range(n_dim))
    if obj is not None:
        for support in get_support(variables, obj):
            nonzeros = np.nonzero(support)[0]
            value = random.random()
            for i in nonzeros:
                for j in nonzeros:
                    rmat[int(i), int(j)] = value
    for polynomial in flatten([inequalities, equalities, momentinequalities,
                               momentequalities]):
        support = np.any(get_support(variables, polynomial), axis=0)
        nonzeros = np.nonzero(support)[0]
        value = random.random()
        for i in nonzeros:
            for j in nonzeros:
                rmat[int(i), int(j)] = value
    rmat = rmat + 5*n_dim*spmatrix(1.0, range(n_dim), range(n_dim))
    # compute symbolic factorization using AMD ordering
    symb = cp.symbolic(rmat, p=amd.order)
    ip = symb.p
    cliques = symb.cliques()
    R = np.zeros((len(cliques), n_dim))
    for i, clique in enumerate(cliques):
        for j in range(len(clique)):
            R[i, ip[cliques[i][j]]] = 1
    return R


def find_clique_index(variables, polynomial, clique_set):
    support = np.any(get_support(variables, polynomial), axis=0)
    support[np.nonzero(support)[0]] = 1
    for i, clique in enumerate(clique_set):
        if np.dot(support, clique) == len(np.nonzero(support)[0]):
            return i
    return -1


def find_variable_cliques(variables, objective=None, inequalities=None,
                          equalities=None, momentinequalities=None,
                          momentequalities=None):
    if objective is None and inequalities is None and equalities is None and \
            momentinequalities is None and momentequalities is None:
        raise Exception("There is nothing to extract the chordal structure " +
                        "from!")
    clique_set = generate_clique(variables, objective, inequalities,
                                 equalities, momentinequalities,
                                 momentequalities)
    variable_sets = []
    for clique in clique_set:
        variable_sets.append([variables[i] for i in np.nonzero(clique)[0]])
    return variable_sets


try:
    from cvxopt import spmatrix, amd
    import chompack as cp
    generate_clique = _generate_clique_alt
except ImportError:
    generate_clique = _generate_clique



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
import abc
from typing import List, Dict, Tuple, Union


class MarginalsAnalyzerInterface(abc.ABC):
    """
    This is interface for classes that will analyze marginal probability distributions.
    It requires those child classes to have basic functionalities that should be included.
    for analyzing marginals_dictionary.
    """

    @property
    @abc.abstractmethod
    def benchmarking_results(self) -> Dict[str, Dict[str, float]]:
        # dictionary of experimental results
        raise NotImplementedError

    @benchmarking_results.setter
    @abc.abstractmethod
    def benchmarking_results(self, benchmarking_results: Dict[str, Dict[str, float]]) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def noisy_marginals_dictionary(self) -> dict:
        # Dictionary storing marginal probability distributions
        raise NotImplementedError


    @noisy_marginals_dictionary.setter
    @abc.abstractmethod
    def noisy_marginals_dictionary(self, marginals_dictionary: dict) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def compute_noisy_marginals(self,
                          experiment_key: str,
                          subsets: List[List[int]]) -> dict:
        """Computes marginals_dictionary for input subsets_list of qubits"""
        raise NotImplementedError

    def compute_hamiltonian_energy_on_classical_state(self,
                                                      weights_dictionary: Dict[Tuple[int], float],
                                                      input_state=Union[str, List[str]]) -> float:
        """Computes energy of diagonal Hamiltonian for given classical input state"""
        raise NotImplementedError



import random
import numpy as np
from typing import List

#TODO TUZJAN Docstrings, typehints

def divide_qubits_in_the_clusters(qubits_indices:List, clusters_specification:List[List])->List[tuple]:
    '''A function used to divide qubits into clusters 

    Parameters: 
    ----------

        qubits_indices: list
            a list of qubits indices defined upon ..... 

        clusters_specification: list
            - a nested list containing description of clusters to be created:

                structure of the list [[element_1, element_2], ...]:
                    element_1 - integer encoding locality of a cluster
                    element_2 - number of clusters of given locality
                
                e.g. [[3,1],[2,2]] encodes 1 3 qubits cluster and two 2 qubits clusters

    Returns
    -------        
    cluster_list : 
    
        list with qubits assigned to clusters

              

    '''
    qubit_indices_now=qubits_indices
    cluster_list = []
    cluster_now =[]
    # a loop over 
    for el in clusters_specification:
        
        #a sublist of tot
        np.random.seed()
        cluster_now=random.sample(list(qubit_indices_now), el[0]*el[1] )
        qubit_indices_now = set(qubit_indices_now) - set(cluster_now)

        cluster_now=[cluster_now[i:i+el[0]] for i in range(0,el[0]*el[1],el[0])]
        for element in cluster_now:
            element.sort()
            cluster_list.append(tuple(element))
    return tuple(cluster_list)



"""
A function creating a string of outcomes from measurement statistics

input parameters:

marginal_results_list - a list with marginal measurement outcomes results are encoded in integers, eg [0,3,0,1,1,2] is a list of 6 results for a two qubit marginals 

"""

def transform_marginal_statistics_to_list_of_results(marginal_results):
    
    results_number= len(marginal_results)
    
    marginal_results_list=[]
    
    for i in range(results_number):
        for j in range(marginal_results[i]):
            marginal_results_list.append(i) 
    random.shuffle(marginal_results_list)

    return marginal_results_list





"""
A function that creates a joint result list by gluing outcomes from two marginals lists 

input parameters:  

marginal_list_1 - a list containing marginal results, results are encoded in integers, eg [0,3,0,1,1,2] is a list of 6 results for a two qubit marginals 

marginal_list_2 - as above 

output parameters:

results_list - a join list of results



"""

'''
A creating a joint sample out of samples created according to product probability distributions

input parameters:  


marginal_results_list - a list of list corresponding to product samples


output parameters:

joint_sample - a list corresponding to the joint sample


'''

def merge_samples_list(marginal_list_1, dim1, marginal_list_2, dim2):
    


    results_list = [0 for i in range(dim1*dim2)]

    for r1, r2 in zip(marginal_list_1,marginal_list_2):
        results_list[r1*dim2+r2]+=1

    return results_list



def merge_results_list(results_1, results_2):

    """

    Function joins marginal results lists. Used in simulation of CN noise model
   
   
   Parameters
    ----------

    results_1: list 
        list with marginal results of an experiment

    results_2: list
        list with marginal results of an experiment
    
    Returns
    ----------
    merged_results
        list with merged results of experiment


    """

    dim_1 = len(results_1)
    dim_2 = len(results_2)
    merged_results = [0 for i in range(dim_1*dim_2)]

    for index_1 in range(dim_1):
        
        for index_2 in range(dim_2):
        
            if results_1[index_1] >= results_2[index_2]:
                element = results_2[index_2]
                merged_results[index_1*dim_1+index_2] = element
                results_1[index_1] = results_1[index_1] - element
                results_2[index_2] = 0
        
            else:
                element = results_1[index_1]
                merged_results[index_1*dim_1+index_2] = element
                results_1[index_1] = 0
                results_2[index_2] = results_2[index_2] - element
                break
    
    return merged_results


def t1():
    list_2 = [80,100]
    list_1 = [60,120]
    print(merge_results_list(list_1,list_2))

if __name__ == "__main__":
    t1()



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
from braket.circuits import Circuit, Gate, Instruction, circuit, Observable
import numpy as np

def _apply_elementary_gate_native(quantum_circuit: Circuit,
                                  qubit_index: int,
                                  gate_label: str)-> Circuit:
    """
    Decomposes quantum gate into elementary rotations in pyquil and adds it to a circuit.

    :param quantum_circuit: Object to which a gate should be added
    :type quantum_circuit: braket.circuits.Circuit

    :param qubit_index: Qubit on which gate should act.
    :type qubit_index: int

    :param gate_label: Symbolic label of the gate. If not supported, returns error.
    Supported gates = ["I", "X", "Y", "Z", "H", "S", "S*", "T", "T*"]
    :type gate_label: str

    :return: "quantum circuit" with a target gate appended.
    :rtype: braket.circuits.Circuit
    """
    if gate_label.upper() == 'I':
        quantum_circuit += Instruction(Gate.I(), qubit_index)
    elif gate_label.upper() == 'X':
        # X
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rx(qubit_index, -np.pi / 2)

    elif gate_label.upper() == 'Y':
        # Y
        quantum_circuit = quantum_circuit.rz(qubit_index, -2.446121240088584)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rx(qubit_index, -np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, -2.446121240088584)

    elif gate_label.upper() == 'Z':
        # Z
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)


    elif gate_label.upper() == 'H':
        # H
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, -np.pi / 2)


    elif gate_label.upper() == 'S':
        # S
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)


    elif gate_label.upper() == 'S*':
        # Sdag
        quantum_circuit = quantum_circuit.rz(qubit_index, -np.pi / 2)

    elif gate_label.upper() == 'T':
        # T

        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 4)

    elif gate_label.upper() == 'T*':
        # Tdag
        quantum_circuit = quantum_circuit.rz(qubit_index, -np.pi / 4)

    else:
        raise ValueError(f"Gate label: '{gate_label}' not recognized.")

    return quantum_circuit


def _apply_pauli_eigenstate(eigenstate_index: int,
                            quantum_circuit: Circuit,
                            qubit_index: int):
    """

    :param eigenstate_index:
    :type eigenstate_index:
    :param quantum_circuit:
    :type quantum_circuit:
    :param qubit_index:
    :type qubit_index:
    :return:
    :rtype:
    """
    # Z+
    if eigenstate_index == 0:
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='I')

    # Z-
    elif eigenstate_index == 1:
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='X')

    # X+
    elif eigenstate_index == 2:
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='H')
    # Y+
    elif eigenstate_index == 3:
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='H')
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='S')

    # X-
    elif eigenstate_index == 4:
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='X')
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='H')
    # Y-
    elif eigenstate_index == 5:
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='X')
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='H')
        quantum_circuit = _apply_elementary_gate_native(quantum_circuit=quantum_circuit,
                                                        qubit_index=qubit_index,
                                                        gate_label='S')

    else:
        raise ValueError(f"Incorrect eigenstate index: '{eigenstate_index}'!")

    return quantum_circuit





"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""

import copy, math, random

import numpy as np
from tqdm import tqdm

from qrem.noise_visualization.functions_data_visualization import *


class InfinityException(Exception):
    """ Class for handling infinite value of cost function  """
    pass


class RejectionException(Exception):
    """ Class for handling rejection of all operations """
    pass


class MaxClusterSizeException(Exception):
    """ Class for handling wrong maximal allowed size of clusters """
    pass


def f_clust_sharp(C_size, C_maxsize):
    """
    Cluster size penalty function f_penalty with sharp threshold.
    f_penalty(C_size) = 0,          if C_size <= C_maxsize,
    f_penalty(C_size) = Infinity,   if C_size > C_maxsize.

    :param C_size: Size of a cluster.
    :type C_size: int

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return: value of the function.
    :rtype: int
    """

    val = 0
    if C_size > C_maxsize:
        raise InfinityException
    return val


def f_clust_square_sharp(C_size, C_maxsize):
    """
    Cluster size penalty function f_penalty with quadratic dependence on C_size and sharp threshold.
    f_penalty(C_size) = (C_size/C_maxsize)^2,   if C_size <= C_maxsize,
    f_penalty(C_size) = Infinity,               if C_size > C_maxsize.

    :param C_size: Size of a cluster.
    :type C_size: int

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return: value of the function.
    :rtype: float
    """

    if C_size > C_maxsize:
        raise InfinityException
    else:
        val = math.pow(C_size, 2)
    return val


def f_clust(C_size, C_maxsize):
    """
    Cluster size penalty function f_penalty with flat beginning, quadratic intermediate values (normalized by C_maxsize) and and sharp threshold.
    f_penalty(C_size) = 0,                      if C_size < 3,
    f_penalty(C_size) = (C_size/C_maxsize)^2,   if 3 <= C_size <= C_maxsize,
    f_penalty(C_size) = Infinity,               if C_size > C_maxsize.

    :param C_size: Size of a cluster.
    :type C_size: int

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return: value of the function.
    :rtype: float
    """

    if C_size > C_maxsize:
        raise InfinityException
    elif C_size < 3:
        val = 0
    else:
        val = math.pow(C_size / C_maxsize, 2)
    return val


def get_Sc_i(C_i, correlations_table):
    """
    Strength of correlaions S_i in cluster i.

    :param C_i: i-th cluster
    :type C_i: List[int]

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :return value of S_i.
    :rtype: float
    """

    if len(C_i) < 2:
        return 0
    val = 0
    for k in C_i:
        for l in C_i:
            if (k == l):
                continue
            val = val + correlations_table[k, l]
    return val


# uses f_clust_square_sharp
def objective_function_simple_cummulative(partition, correlations_table, alpha, C_maxsize):
    """
    Calculates the value of the objective function phi on partition.

    :param partition: partition
    :type partition: List[int]

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return value of phi on partition.
    :rtype: float
    """

    no_of_clusters = len(partition)
    val = 0
    for C_i in partition:
        try:
            val = val - alpha * f_clust_square_sharp(len(C_i), C_maxsize)
        except InfinityException:
            raise InfinityException

    for i in range(no_of_clusters):
        C_i = partition[i]
        S_i = get_Sc_i(C_i, correlations_table)
        val = val + S_i
    return val


# uses f_clust_square_sharp
def evaluate_move_operation_naive_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha,
                                              C_maxsize):
    """
    Calculates the change of the objective function phi after specified move operation. Calculations are naive - objective function is calculated twice and the difference is taken. Can be used for debugging.

    :param partition: partition
    :type partition: List[int]

    :param index_k: qubit to be moved
    :type index_k: int

    :param index_C_i: index of a cluster with qubit index_k
    :type index_C_i: int

    :param index_C_j: index of a cluster where index_k is to be moved
    :type index_C_j: int

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return change of phi under operation, if the value is finite
    :rtype:(float, bool)
    """

    partition_copy = copy.deepcopy(partition)
    val1 = objective_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    make_move_operation(partition_copy, index_k, index_C_i, index_C_j)
    try:
        val2 = objective_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    except InfinityException:
        return 0, False
    diff = val2 - val1
    return diff, True


# uses f_clust_square_sharp
def evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j, correlations_table, alpha, C_maxsize):
    """
    Calculates the change of the objective function phi after specified move operation.

    :param partition: partition
    :type partition: List[int]

    :param index_k: qubit to be moved
    :type index_k: int

    :param index_C_i: index of a cluster with qubit index_k
    :type index_C_i: int

    :param index_C_j: index of a cluster where index_k is to be moved
    :type index_C_j: int

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return change of phi under operation, if the value is finite
    :rtype:(float, bool)
    """

    C_i = partition[index_C_i]
    C_j = partition[index_C_j]
    c_i = len(C_i)
    c_j = len(C_j)

    diff = 0
    # cluster function part
    try:
        dS_clust = f_clust_square_sharp(c_i, C_maxsize) + f_clust_square_sharp(c_j, C_maxsize) - f_clust_square_sharp(
            c_i - 1, C_maxsize) - f_clust_square_sharp(c_j + 1, C_maxsize)
        dS_clust = alpha * dS_clust
    except InfinityException:
        return diff, False

    # intra cluster part
    dS_intra = 0
    for index_q in C_j:
        dS_intra = dS_intra + correlations_table[index_q][index_k] + correlations_table[index_k][index_q]
    for index_q in C_i:
        if (index_q == index_k):
            continue
        dS_intra = dS_intra - correlations_table[index_q][index_k] - correlations_table[index_k][index_q]

    diff = dS_clust + dS_intra
    return diff, True


# uses f_clust_square_sharp
def evaluate_swap_operation_naive_cummulative(partition, index_k, index_l, index_C_i, index_C_j, correlations_table,
                                              alpha, C_maxsize):
    """
    Calculates the change of the objective function phi after specified swap operation. Calculations are naive - objective function is calculated twice and the difference is taken. Can be used for debugging.

    :param partition: partition
    :type partition: List[int]

    :param index_k: qubit to be swapped
    :type index_k: int

    :param index_l: qubit to be swapped
    :type index_l: int

    :param index_C_i: index of a cluster with qubit index_k
    :type index_C_i: int

    :param index_C_j: index of a cluste with qubit index_l
    :type index_C_j: int

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :return change of phi under operation, if the value is finite
    :rtype:(float, bool)
    """

    partition_copy = copy.deepcopy(partition)
    val1 = objective_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    make_swap_operation(partition_copy, index_k, index_l, index_C_i, index_C_j)
    val2 = objective_function_simple_cummulative(partition_copy, correlations_table, alpha, C_maxsize)
    diff = val2 - val1
    return diff


def evaluate_swap_operation_cummulative(partition, index_k, index_l, index_C_i, index_C_j, correlations_table, alpha):
    """
   Calculates the change of the objective function phi after specified swap operation.

   :param partition: partition
   :type partition: List[int]

   :param index_k: qubit to be swapped
   :type index_k: int

   :param index_l: qubit to be swapped
   :type index_l: int

   :param index_C_i: index of a cluster with qubit index_k
   :type index_C_i: int

   :param index_C_j: index of a cluste with qubit index_l
   :type index_C_j: int

   :param correlations_table: table of all 2-qubit correlation coefficients.
   :type correlations_table: ndarray(dtype=float, ndim=2)

   :param alpha: multiplicative parameter of the cluster size penalty function.
   :type alpha: float

   :param C_maxsize: Maximal allowed size of a cluster (threshold).
   :type C_maxsize: int

   :return change of phi under operation, if the value is finite
   :rtype:(float, bool)
   """

    C_i = partition[index_C_i]
    C_j = partition[index_C_j]

    diff = 0
    # intra cluster part
    dS_intra = 0
    for index_q in C_i:
        if (index_q == index_k):
            continue
        dS_intra = dS_intra + correlations_table[index_q][index_l] + correlations_table[index_l][index_q] - \
                   correlations_table[index_q][index_k] - correlations_table[index_k][index_q]
    for index_q in C_j:
        if (index_q == index_l):
            continue
        dS_intra = dS_intra + correlations_table[index_q][index_k] + correlations_table[index_k][index_q] - \
                   correlations_table[index_q][index_l] - correlations_table[index_l][index_q]
        diff = dS_intra
    return diff


def get_initial_partition(correlations_table):
    """
    Finds the initial partition based on thr correlatons table. Pairs of qubits with largest correlations are put to the same 2-qubit cluster. If number of qubits is odd, one qubit remains unpaired.

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :return: partition
    :rtype: List[List[int]]
    """

    n = correlations_table.shape[0]
    CL = dict()
    for i in range(n - 1):
        for j in range(i + 1, n):
            val1 = correlations_table[i, j]
            val2 = correlations_table[j, i]
            if (val1 > val2):
                val = val1
            else:
                val = val2
            CL.update({(i, j): val})
    partition = []
    qubits = list(range(n))
    while len(qubits) > 1:
        pair = max(CL, key=lambda key: CL[key])
        i = pair[0]
        j = pair[1]
        partition.append([i, j])
        keys = list(CL.keys())
        for pair in keys:
            if pair[0] in [i, j] or pair[1] in [i, j]:
                CL.pop(pair, None)
        qubits.remove(i)
        qubits.remove(j)
    if len(qubits) == 1:
        partition.append([qubits[0]])

    return partition


def return_cluster_index(partition, target_qubit):
    """
    Finds the index of a cluster in a partition in which is the target_qubit.

    :param partition: list of clusters
    :type partition: List[List[int]]

    :param target_qubit: qubit whose cluster's index we are looking
    :type target_qubit: int

    :return: cluster index in a partition.
    :rtype: int
    """

    index = 0
    for cluster in partition:
        for qubit in cluster:
            if qubit == target_qubit:
                return index
        index = index + 1
    return index


def make_move_operation(partition, index_k, index_C_i, index_C_j):
    """
    Realizes the move operation on partition, where qubit index_k is moved from cluster with index index_C_i to cluster with index index_C_j.

    :param partition: list of clusters
    :type partition: List[List[int]]

    :param index_k: qubit to be moved
    :type index_k: int

    :param index_C_i: index of a cluster with qubit index_k
    :type index_C_i: int

    :param index_C_j: index of a cluster where index_k is to be moved
    :type index_C_j: int
    """
    partition[index_C_j].append(index_k)

    if len(partition[index_C_i]) == 1:
        partition.pop(index_C_i)
    else:
        partition[index_C_i].remove(index_k)


def make_swap_operation(partition, index_k, index_l, index_C_i, index_C_j):
    """
    Realizes the swap operation on partition, where qubit index_k from cluster with index index_C_i is swapped with qubit index_l from cluster with index index_C_j

    :param partition: list of clusters
    :type partition: List[List[int]]

    :param index_k: qubit to be swapped
    :type index_k: int

    :param index_l: qubit to be swapped
    :type index_l: int

    :param index_C_i: index of a cluster with qubit index_k
    :type index_C_i: int

    :param index_C_j: index of a cluste with qubit index_l
    :type index_C_j: int
    """

    partition[index_C_i].remove(index_k)
    partition[index_C_i].append(index_l)
    partition[index_C_j].remove(index_l)
    partition[index_C_j].append(index_k)


def choose_pos_op(val_of_ops_pos, temp):
    """
    Chooses the cluster index for which to move the qubit among all those resulting in positive change of objective function phi.
    Probability of the choice of cluster i is:
    p(i)=exp((change_i-max_change)/temp)/norm_const
    where norm_const is the normalization constant: sum_j exp((change_j-max_change)/temp.

    :param val_of_ops_pos: dictionary with keys being cluster indices and values being the changes of phi.
    :type val_of_ops_pos: dict

    :param temp: temperature parameter
    :type temp: float

    :return: chosen cluster index.
    :rtype: int
    """

    prob_dict = dict()
    values = np.array(list(val_of_ops_pos.values()))
    max_value = np.max(values)
    exp_args = (values - max_value) / temp
    exp_vals = np.exp(exp_args)
    norm_const = np.sum(exp_vals)
    for op in val_of_ops_pos:
        prob_dict[op] = np.exp((val_of_ops_pos[op] - max_value) / temp) / norm_const
    chosen_op = random.choices(list(prob_dict.keys()), weights=prob_dict.values(), k=1)[0]
    return chosen_op


def choose_nonpos_op(val_of_ops_nonpos, temp):
    """
    Chooses the cluster index for which to move the qubit among all those resulting in nonpositive change of objective function phi.
    Probability of the choice of cluster i is:
    p(i)=exp(change_i/temp)/(no_of_ops_nonpos + epsilon)
    where no_of_ops_nonpos is the number of all operations and epsilon is small number added for numerical stability (here set as 0.001).
    Additionally, no operation can be selected with probability:
    p(reject)= 1- sum_j p(j).

    :param val_of_ops_nonpos: dictionary with keys being cluster indices and values being the changes of phi.
    :type val_of_ops_nonpos: dict

    :param temp: temperature parameter
    :type temp: float

    :return: chosen cluster index.
    :rtype: int
    """

    prob_dict = dict()
    no_of_ops_nonpos = len(val_of_ops_nonpos.keys())
    psum = 0
    for op in val_of_ops_nonpos:
        p = np.exp(val_of_ops_nonpos[op] / temp) / (no_of_ops_nonpos + 0.001)
        prob_dict[op] = p
        psum += p
    prob_dict['reject'] = 1.0 - psum
    chosen_op = random.choices(list(prob_dict.keys()), weights=prob_dict.values(), k=1)[0]
    if chosen_op == 'reject':
        raise RejectionException
    return chosen_op


# uses f_clust_square_sharp, objective_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
def partition_algorithm_v1_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       printing,
                                       drawing, disable_pb):
    """
    Partition algorithm using move and swap operations on pairs of qubits with termination criterion.
    If C_maxsize < 2 then algorithm terminates with exception raised. If C_maxsize = 2 then the algorithm is not executed and the initial partition is returned.
    The algorith is run N_alg times and the best partition is chosen. In each run, the algorithm traverses pairs of qubits in epochs.
    An epoch consists of all pairs of qubits that are traversed at random and without repetitions.
    The partition is being updated during each epoch, by executing move and swap operations on pairs of qubits.
    The algorithm terminates, if after some epoch the value of objective function phi has not improved compared to the end of a previous epoch.

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :param N_alg: Number of runs.
    :type N_alg: int

    :param printing: Is printing allowed.
    :type printing: bool

    :param drawing: Is cluster drawing allowed.
    :type drawing: bool

    :param disable_pb: Is progress bar disabled.
    :type disable_pb: bool

    :return best partition found, value of the objective function for this partition
    :rtype:(List[int], float)

    :raises MaxClusterSizeException: returned if max allowed cluster size is < 2
    """

    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        raise MaxClusterSizeException
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = objective_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    global_best_parition = initial_partition
    global_best_value = init_cf
    if (skip_optimisation == False):
        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf - 1
            curr_cf = init_cf
            epoch_no = 0
            while curr_cf > best_cf:
                epoch_no = epoch_no + 1
                best_cf = curr_cf
                if (printing):
                    print("starting epoch: " + str(epoch_no))
                pairs = []
                for i in range(no_of_qubits - 1):
                    for j in range(i + 1, no_of_qubits):
                        pairs.append([i, j])
                while len(pairs) > 0:
                    val_of_ops = dict()
                    pair = random.choice(pairs)
                    pairs.remove(pair)
                    i = pair[0]
                    j = pair[1]
                    if (printing):
                        print("pair: " + "(" + str(i) + ", " + str(j) + ")")
                    index_C_i = return_cluster_index(partition, i)
                    index_C_j = return_cluster_index(partition, j)
                    if (index_C_i == index_C_j):
                        if (printing):
                            print("WRONG PAIR - SKIP")
                        continue
                    index_k = i
                    val1, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                              correlations_table, alpha, C_maxsize)

                    if (printing):
                        print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val1) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val1 > 0 and not_neg_infty:
                        val_of_ops.update({'move_ij': val1})
                    index_C_i = return_cluster_index(partition, j)
                    index_C_j = return_cluster_index(partition, i)
                    index_k = j
                    val2, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                              correlations_table, alpha, C_maxsize)
                    if (printing):
                        print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val2) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val2 > 0 and not_neg_infty:
                        val_of_ops.update({'move_ji': val2})
                    index_C_i = return_cluster_index(partition, i)
                    index_C_j = return_cluster_index(partition, j)
                    index_k = i
                    index_l = j
                    val3 = evaluate_swap_operation_cummulative(partition, index_k, index_l, index_C_i, index_C_j,
                                                               correlations_table, alpha)
                    if (printing):
                        print("checking swap " + str(index_C_i) + "<-" + str(index_k) + "&" + str(index_l) + "->" + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val3) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val3 > 0:
                        val_of_ops.update({'swap': val3})

                    if len(val_of_ops) > 0:
                        if (printing):
                            print("ACCEPT:")
                        op = max(val_of_ops, key=lambda key: val_of_ops[key])
                        if (op == 'move_ij'):
                            index_C_i = return_cluster_index(partition, i)
                            index_C_j = return_cluster_index(partition, j)
                            index_k = i
                            make_move_operation(partition, index_k, index_C_i, index_C_j)
                            if (printing):
                                print("MOVE_ij")
                                print(partition)
                            curr_cf = curr_cf + val_of_ops[op]
                        elif (op == 'move_ji'):
                            index_C_i = return_cluster_index(partition, j)
                            index_C_j = return_cluster_index(partition, i)
                            index_k = j
                            make_move_operation(partition, index_k, index_C_i, index_C_j)
                            if (printing):
                                print("MOVE_ji")
                                print(partition)
                            curr_cf = curr_cf + val_of_ops[op]
                        else:
                            index_C_i = return_cluster_index(partition, i)
                            index_C_j = return_cluster_index(partition, j)
                            index_k = i
                            index_l = j
                            make_swap_operation(partition, index_k, index_l, index_C_i, index_C_j)
                            if (printing):
                                print("SWAP")
                                print(partition)
                            curr_cf = curr_cf + val_of_ops[op]
                    elif (printing):
                        print("REJECT ALL")
            if (printing):
                print("convergence to: " + str(best_cf) + " after " + str(epoch_no) + " epochs")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value


# TODO OS: add optional convergence detection (termination after N_rand steps or if convergence detected). For example if a pair is rejected it goes to a list. If pair is accepted the list is erased. If list is full the algorithm terminates current attempt.
# uses f_clust_square_sharp, objective_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
def partition_algorithm_v2_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       N_rand,
                                       printing,
                                       drawing, disable_pb):
    """
    Partition algorithm using move and swap operations on pairs of qubits at random.
    If C_maxsize < 2 then algorithm terminates with exception raised. If C_maxsize = 2 then the algorithm is not executed and the initial partition is returned.
    The algorith is run N_alg times and the best partition is chosen. In each run, the algorithm traverses pairs of qubits.
    Pairs of qubits are being traversed at random with repetitions and a total of N_rand pairs is chosen.
    The partition is being updated by executing move and swap operations on pairs of qubits.
    The algorithm terminates after N_runs. Each run terminates after considering N_rand pairs.

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :param N_alg: Number of runs.
    :type N_alg: int
    
    :param N_rand: Number of pairs to be chosen at random on each run.
    :type N_rand: int

    :param printing: Is printing allowed.
    :type printing: bool

    :param drawing: Is cluster drawing allowed.
    :type drawing: bool

    :param disable_pb: Is progress bar disabled.
    :type disable_pb: bool

    :return best partition found, value of the objective function for this partition
    :rtype:(List[int], float)

    :raises MaxClusterSizeException: returned if max allowed cluster size is < 2
    """

    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        raise MaxClusterSizeException
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = objective_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    global_best_parition = initial_partition
    global_best_value = init_cf

    if (skip_optimisation == False):

        pairs = []
        for i in range(no_of_qubits - 1):
            for j in range(i + 1, no_of_qubits):
                pairs.append([i, j])

        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf
            move_no = 0

            while move_no <= N_rand - 1:
                move_no = move_no + 1
                if (printing):
                    print("starting move: " + str(move_no))

                val_of_ops = dict()
                pair = random.choice(pairs)
                i = pair[0]
                j = pair[1]
                if (printing):
                    print("pair: " + "(" + str(i) + ", " + str(j) + ")")
                index_C_i = return_cluster_index(partition, i)
                index_C_j = return_cluster_index(partition, j)
                if (index_C_i == index_C_j):
                    if (printing):
                        print("WRONG PAIR - SKIP")
                    continue
                index_k = i
                val1, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                          correlations_table, alpha, C_maxsize)

                if (printing):
                    print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val1) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val1 > 0 and not_neg_infty:
                    val_of_ops.update({'move_ij': val1})
                index_C_i = return_cluster_index(partition, j)
                index_C_j = return_cluster_index(partition, i)
                index_k = j
                val2, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                          correlations_table, alpha, C_maxsize)
                if (printing):
                    print("checking move " + str(index_C_i) + "--" + str(index_k) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val2) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val2 > 0 and not_neg_infty:
                    val_of_ops.update({'move_ji': val2})
                index_C_i = return_cluster_index(partition, i)
                index_C_j = return_cluster_index(partition, j)
                index_k = i
                index_l = j
                val3 = evaluate_swap_operation_cummulative(partition, index_k, index_l, index_C_i, index_C_j,
                                                           correlations_table, alpha)
                if (printing):
                    print("checking swap " + str(index_C_i) + "<-" + str(index_k) + "&" + str(index_l) + "->" + str(
                        index_C_j) + "\t, result: " + "{:.5f}".format(val3) + "\t, size_leq_C_max: " + str(
                        not_neg_infty))
                if val3 > 0:
                    val_of_ops.update({'swap': val3})

                if len(val_of_ops) > 0:
                    if (printing):
                        print("ACCEPT:")
                    op = max(val_of_ops, key=lambda key: val_of_ops[key])
                    if (op == 'move_ij'):
                        index_C_i = return_cluster_index(partition, i)
                        index_C_j = return_cluster_index(partition, j)
                        index_k = i
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("MOVE_ij")
                            print(partition)
                        best_cf = best_cf + val_of_ops[op]
                    elif (op == 'move_ji'):
                        index_C_i = return_cluster_index(partition, j)
                        index_C_j = return_cluster_index(partition, i)
                        index_k = j
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("MOVE_ji")
                            print(partition)
                        best_cf = best_cf + val_of_ops[op]
                    else:
                        index_C_i = return_cluster_index(partition, i)
                        index_C_j = return_cluster_index(partition, j)
                        index_k = i
                        index_l = j
                        make_swap_operation(partition, index_k, index_l, index_C_i, index_C_j)
                        if (printing):
                            print("SWAP")
                            print(partition)
                        best_cf = best_cf + val_of_ops[op]
                elif (printing):
                    print("REJECT ALL")
            if (printing):
                print("Obtained: " + str(best_cf) + " after " + str(move_no) + " moves")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value


# TODO OS: add optional convergence detection (termination after N_rand steps or if convergence detected)
# uses f_clust_square_sharp, objective_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
def partition_algorithm_v3_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       N_rand,
                                       printing,
                                       drawing, disable_pb):
    """
   Partition algorithm using move operations on qubits at random.
   If C_maxsize < 2 then algorithm terminates with exception raised. If C_maxsize = 2 then the algorithm is not executed and the initial partition is returned.
   The algorith is run N_alg times and the best partition is chosen. In each run, the algorithm traverses qubits.
   Qubits are being traversed at random with repetitions and a total of N_rand qubits chosen.
   The partition is being updated by executing move operations on qubits.
   The algorithm terminates after N_runs. Each run terminates after considering N_rand qubits.

   :param correlations_table: table of all 2-qubit correlation coefficients.
   :type correlations_table: ndarray(dtype=float, ndim=2)

   :param alpha: multiplicative parameter of the cluster size penalty function.
   :type alpha: float

   :param C_maxsize: Maximal allowed size of a cluster (threshold).
   :type C_maxsize: int

   :param N_alg: Number of runs.
   :type N_alg: int

   :param N_rand: Number of qubits to be chosen at random on each run.
   :type N_rand: int

   :param printing: Is printing allowed.
   :type printing: bool

   :param drawing: Is cluster drawing allowed.
   :type drawing: bool

   :param disable_pb: Is progress bar disabled.
   :type disable_pb: bool

   :return best partition found, value of the objective function for this partition
   :rtype:(List[int], float)

   :raises MaxClusterSizeException: returned if max allowed cluster size is < 2
   """

    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        raise MaxClusterSizeException
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = objective_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    global_best_parition = initial_partition
    global_best_value = init_cf

    if (skip_optimisation == False):

        qubits = range(no_of_qubits)

        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf
            move_no = 0

            while move_no <= N_rand - 1:
                move_no = move_no + 1
                if (printing):
                    print("starting move: " + str(move_no))

                val_of_ops = dict()
                i = random.choice(qubits)
                if (printing):
                    print("selected qubit: " + str(i))
                index_C_i = return_cluster_index(partition, i)
                index_k = i
                for index_C_j in range(len(partition)):
                    if index_C_j == index_C_i:
                        continue

                    val, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i, index_C_j,
                                                                             correlations_table, alpha, C_maxsize)

                    if (printing):
                        print("checking move of qubit " + str(i) + " to partition " + str(
                            index_C_j) + "\t, result: " + "{:.5f}".format(val) + "\t, size_leq_C_max: " + str(
                            not_neg_infty))
                    if val > 0 and not_neg_infty:
                        val_of_ops.update({index_C_j: val})
                if len(val_of_ops) > 0:
                    best_cluster_index = max(val_of_ops, key=lambda key: val_of_ops[key])
                    best_cf = best_cf + val_of_ops[best_cluster_index]
                    index_C_j = best_cluster_index
                    make_move_operation(partition, index_k, index_C_i, index_C_j)
                    if (printing):
                        print("ACCEPT move to cluster no " + str(best_cluster_index) + ", value=" + str(best_cf))
                        print(partition)
                elif (printing):
                    print("REJECT ALL")

            if (printing):
                print("Obtained: " + str(best_cf) + " after move " + str(move_no))
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value


# uses f_clust_square_sharp, objective_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
def partition_algorithm_v4_cummulative(correlations_table,
                                       alpha,
                                       C_maxsize,
                                       N_alg,
                                       printing,
                                       drawing, disable_pb):
    """
    Partition algorithm using move and swap operations on qubits with termination criterion.
    If C_maxsize < 2 then algorithm terminates with exception raised. If C_maxsize = 2 then the algorithm is not executed and the initial partition is returned.
    The algorith is run N_alg times and the best partition is chosen. In each run, the algorithm traverses qubits in epochs.
    An epoch consists of all qubits that are traversed at random and without repetitions.
    The partition is being updated during each epoch, by executing move and swap operations on qubits.
    The algorithm terminates, if after some epoch the value of objective function phi has not improved compared to the end of a previous epoch.

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :param N_alg: Number of runs.
    :type N_alg: int

    :param printing: Is printing allowed.
    :type printing: bool

    :param drawing: Is cluster drawing allowed.
    :type drawing: bool

    :param disable_pb: Is progress bar disabled.
    :type disable_pb: bool

    :return best partition found, value of the objective function for this partition
    :rtype:(List[int], float)

    :raises MaxClusterSizeException: returned if max allowed cluster size is < 2
    """

    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        raise MaxClusterSizeException
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = objective_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    global_best_parition = initial_partition
    global_best_value = init_cf
    if (skip_optimisation == False):
        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf - 1
            curr_cf = init_cf
            epoch_no = 0
            while curr_cf > best_cf:
                epoch_no = epoch_no + 1
                best_cf = curr_cf
                if (printing):
                    print("starting epoch: " + str(epoch_no))
                qubits = list(range(no_of_qubits))

                while len(qubits) > 0:
                    val_of_ops = dict()
                    i = random.choice(qubits)
                    qubits.remove(i)

                    if (printing):
                        print("selected qubit: " + str(i))
                    index_C_i = return_cluster_index(partition, i)
                    index_k = i
                    for index_C_j in range(len(partition)):
                        if index_C_j == index_C_i:
                            continue

                        val, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i,
                                                                                 index_C_j, correlations_table, alpha,
                                                                                 C_maxsize)

                        if (printing):
                            print("checking move of qubit " + str(i) + " to partition " + str(
                                index_C_j) + "\t, result: " + "{:.5f}".format(val) + "\t, size_leq_C_max: " + str(
                                not_neg_infty))
                        if val > 0 and not_neg_infty:
                            val_of_ops.update({index_C_j: val})
                    if len(val_of_ops) > 0:
                        best_cluster_index = max(val_of_ops, key=lambda key: val_of_ops[key])
                        curr_cf = curr_cf + val_of_ops[best_cluster_index]
                        index_C_j = best_cluster_index
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("ACCEPT move to cluster no " + str(best_cluster_index) + ", value=" + str(curr_cf))
                            print(partition)
                    elif (printing):
                        print("REJECT ALL")

            if (printing):
                print("convergence to: " + str(best_cf) + " after " + str(epoch_no) + " epochs")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value


# uses f_clust_square_sharp, objective_function_simple_cummulative, and non-naive move and swap cummulative evaluation operations
def partition_algorithm_v4_cummulative_temp(correlations_table,
                                            alpha,
                                            C_maxsize,
                                            N_alg,
                                            printing,
                                            drawing, disable_pb, temp=1.0):
    """
    Partition algorithm using move operations on qubits with termination criterion and metropolis-like acceptance condition (regulated by temp).
    If C_maxsize < 2 then algorithm terminates with exception raised. If C_maxsize = 2 then the algorithm is not executed and the initial partition is returned.
    The algorith is run N_alg times and the best partition is chosen. In each run, the algorithm traverses qubits in epochs.
    An epoch consists of all qubits that are traversed at random and without repetitions.
    The partition is being updated during each epoch, by executing move operations on qubits.
    The algorithm terminates, if after some epoch the value of objective function phi has not improved compared to the end of a previous epoch.

    :param correlations_table: table of all 2-qubit correlation coefficients.
    :type correlations_table: ndarray(dtype=float, ndim=2)

    :param alpha: multiplicative parameter of the cluster size penalty function.
    :type alpha: float

    :param C_maxsize: Maximal allowed size of a cluster (threshold).
    :type C_maxsize: int

    :param N_alg: Number of runs.
    :type N_alg: int

    :param printing: Is printing allowed.
    :type printing: bool

    :param drawing: Is cluster drawing allowed.
    :type drawing: bool

    :param disable_pb: Is progress bar disabled.
    :type disable_pb: bool

    :param temp: Temperture parameter (default is 1.0).
    :type temp: float

    :return best partition found, value of the objective function for this partition
    :rtype:(List[int], float)

    :raises MaxClusterSizeException: returned if max allowed cluster size is < 2
    """

    skip_optimisation = False

    if (C_maxsize < 2):
        print("Error: max cluster size has to be at least 2!. Algorithm terminated.")
        raise MaxClusterSizeException
    elif (C_maxsize == 2):
        skip_optimisation = True
    initial_partition = get_initial_partition(correlations_table)
    if (printing):
        print("initial partition:")
        print(initial_partition)
    if (drawing):
        print_partition(initial_partition)
    init_cf = objective_function_simple_cummulative(initial_partition, correlations_table, alpha, C_maxsize)
    if (printing):
        print("initial value: " + str(init_cf))
    no_of_qubits = correlations_table.shape[0]
    global_best_parition = initial_partition
    global_best_value = init_cf
    if (skip_optimisation == False):
        for attempt_no in tqdm(range(N_alg), disable=disable_pb):
            if (printing):
                print("attempt: " + str(attempt_no + 1))
            partition = copy.deepcopy(initial_partition)
            best_cf = init_cf - 1
            curr_cf = init_cf
            epoch_no = 0
            while curr_cf > best_cf:
                epoch_no = epoch_no + 1
                best_cf = curr_cf
                if (printing):
                    print("starting epoch: " + str(epoch_no))
                qubits = list(range(no_of_qubits))

                while len(qubits) > 0:
                    val_of_ops_pos = dict()
                    val_of_ops_nonpos = dict()
                    i = random.choice(qubits)
                    qubits.remove(i)

                    if (printing):
                        print("selected qubit: " + str(i))
                    index_C_i = return_cluster_index(partition, i)
                    index_k = i
                    for index_C_j in range(len(partition)):
                        if index_C_j == index_C_i:
                            continue

                        val, not_neg_infty = evaluate_move_operation_cummulative(partition, index_k, index_C_i,
                                                                                 index_C_j, correlations_table, alpha,
                                                                                 C_maxsize)

                        if (printing):
                            print("checking move of qubit " + str(i) + " to partition " + str(
                                index_C_j) + "\t, result: " + "{:.5f}".format(val) + "\t, size_leq_C_max: " + str(
                                not_neg_infty))
                        if val > 0 and not_neg_infty:
                            val_of_ops_pos.update({index_C_j: val})
                        elif val <= 0 and not_neg_infty:
                            val_of_ops_nonpos.update({index_C_j: val})
                    if len(val_of_ops_pos) > 0:
                        chosen_cluster_index = choose_pos_op(val_of_ops_pos, temp)
                        curr_cf = curr_cf + val_of_ops_pos[chosen_cluster_index]
                        index_C_j = chosen_cluster_index
                        make_move_operation(partition, index_k, index_C_i, index_C_j)
                        if (printing):
                            print("ACCEPT move to cluster no " + str(chosen_cluster_index) + ", value=" + str(curr_cf))
                            print(partition)
                    else:
                        try:
                            chosen_cluster_index = choose_nonpos_op(val_of_ops_nonpos, temp)
                            curr_cf = curr_cf + val_of_ops_nonpos[chosen_cluster_index]
                            index_C_j = chosen_cluster_index
                            make_move_operation(partition, index_k, index_C_i, index_C_j)
                            if (printing):
                                print("ACCEPT move to cluster no " + str(chosen_cluster_index) + ", value=" + str(
                                    curr_cf))
                                print(partition)
                        except RejectionException:
                            if (printing):
                                print("REJECT ALL")

            if (printing):
                print("convergence to: " + str(best_cf) + " after " + str(epoch_no) + " epochs")
            if (best_cf > global_best_value):
                global_best_parition = partition
                global_best_value = best_cf

    global_best_parition_sorted = [sorted(cluster) for cluster in global_best_parition]
    return global_best_parition_sorted, global_best_value



from typing import Optional, Dict, List, Union, Tuple

from tqdm import tqdm
import numpy as np

from qrem.noise_characterization.data_analysis.InitialNoiseAnalyzer import InitialNoiseAnalyzer
from qrem.functions_qrem import ancillary_functions as anf, povmtools
from qrem.noise_characterization.tomography_design.standard.QuantumDetectorTomography import QDTCalibrationSetup, QDTFitter

from qrem.common.math import Constants as Const
from qrem.common.math import ketbra
"""
JT: This class is used to:
    - perform computations of marginals
    - compute reduced POVMs


Analysis    

"""


#TODO FBM: this should probably descend from different class?
class QDTMarginalsAnalyzer(InitialNoiseAnalyzer):
    def __init__(self,
                 results_dictionary: Dict[str, Dict[str, int]],
                 experiment_name: str,
                 marginals_dictionary: Optional[Dict[str, Dict[str, np.ndarray]]] = None,
                 marginals_averaged: Optional[Dict[str, Dict[str, np.ndarray]]] = None,
                 POVM_dictionary: Optional[Dict[str, List[np.ndarray]]] = None,
                 convergence_treshold_QDT: Optional[float] = 10 ** (-6)) -> None:

        super().__init__(results_dictionary=results_dictionary,
                         marginals_dictionary=marginals_dictionary
                         )

        if marginals_averaged is None:
            marginals_averaged = {}
        if POVM_dictionary is None:
            POVM_dictionary = {}

        self._experiment_name = experiment_name

        self._labels_interpreter = None

        self._convergence_treshold_QDT = convergence_treshold_QDT



        self._marginals_averaged = marginals_averaged

    @property
    def POVM_dictionary(self) -> Dict[Tuple[int], List[np.ndarray]]:
        return self._POVM_dictionary

    @property
    def marginals_averaged(self) -> Dict[Tuple[int], Dict[Tuple[int], np.ndarray]]:
        return self._marginals_averaged
#JT: labels are used to interpret abstract symbols used to create tomographic circuits in terms of physical states used in tomography
#By Pauli an overcomplete bais of 6 eigenstates of Pauli eigenstates is ment
    @staticmethod
    def _labels_interpreter_paulis(label_gate):
        #TODO FBM: TESTS
        _pauli_labels = ['z+', 'z-', 'x+', 'x-', 'y+', 'y-']
        # _pauli_labels = ['z+', 'z-', 'x+', 'y+', 'x-', 'y-']

        if label_gate not in ['%s' % i for i in range(6)] + list(range(6)):
            raise ValueError('Wrong label: ' + label_gate + ' for Pauli eigentket.')
        #JT: here a dictionary mapping integers to states labels is performed
        mapping = {i: _pauli_labels[i] for i in range(6)}
        #JT: This line extends the above dictionary so that it maps symbols to states and states to symbols so that we have 0: 'z+' and 'z+':0
        mapping = {**mapping, **{'%s' % i: _pauli_labels[i] for i in range(6)}}

        #JT: This loop in this place is redundant, for now I comment it
        """for i in range(6):
            mapping[i] = _pauli_labels[i]"""
        #The labels correspoding to Pauli eigenstates are interpreted as states, __pauli_eigenkets returns vectors corresponding to normalized eigenstates
        return Const.pauli_eigenkets()[mapping[label_gate]].reshape(2,1)

#JT: The same as above, the only difference is ordering of Pauli labels
    @staticmethod
    def __labels_interpreter_paulis_rigetti_debug(label_gate):
        _pauli_labels_rigetti_debug = ['z+', 'z-', 'x+', 'y+', 'x-', 'y-']

        if label_gate not in ['%s' % i for i in range(6)] + list(range(6)):
            raise ValueError('Wrong label: ' + label_gate + ' for Pauli eigentket.')

        mapping = {i: _pauli_labels_rigetti_debug[i] for i in range(6)}
        mapping = {**mapping, **{'%s' % i: _pauli_labels_rigetti_debug[i] for i in range(6)}}
        for i in range(6):
            mapping[i] = _pauli_labels_rigetti_debug[i]

        return Const.pauli_eigenkets()[mapping[label_gate]].reshape(2,1)

    """
    JT: initialize_labels_interpreter: sets interpreter of abstract symbols used in circutis creaation to the states realized in the experiments
    for PAULI we interpret symbols as 6 eigenstates of the Pauli operators
    For some reason for some older experiments on Rigetti the order of the eigenstates was different that that on IBM
    or AWS implemented by Asia. To take this into account there is an additional bool parameter rigetti_debug, which for
    analysis of Rigetii experimental data form March must be set to True, otherwise it should be false      
    """


    def initialize_labels_interpreter(self,
                                      interpreter,
                                      rigetti_debug=False):
        if isinstance(interpreter, str):
            if interpreter.upper() in ['PAULI', 'PAULIS']:
                #JT: here an interpreter is set, see above
                if not rigetti_debug:
                    interpreter = self._labels_interpreter_paulis
                else:
                    interpreter = self.__labels_interpreter_paulis_rigetti_debug
            else:
                raise ValueError('Basis: ' + interpreter + ' not yet supported. \
                    Please provide custom interpreter.')

        self._labels_interpreter = interpreter




#JT: a function used to create a multiqubit state out of a string of integers, uses interpreter. It is used in POVMs computations via ML reconstruction
    def _get_tensored_ket(self,
                          key_multiple):
        ket_now = 1
        for key_state in key_multiple:
            ket_now = np.kron(ket_now,self._labels_interpreter(key_state))
        return ket_now

#JT: a function used to create a multiqubit density matrix out of a string of integers, uses interpreter. It is used in POVMs computations via PLS reconstruction
#the terms 3 * projector onto a state and - identity matrix stem from formula for PLS reconstruction

    def _get_tensored_ket_LS(self,
                          key_multiple):
        ket_now = 1
        for key_state in key_multiple:
            ket_now = np.kron(ket_now,
                              3*ketbra(self._labels_interpreter(key_state))-np.eye(2,
                                                                           dtype=complex))
        return ket_now


    def _compute_averaged_marginals_subset(self,
                                           subset:Tuple[int]):
        # subset_key = 'q' + 'q'.join([str(s) for s in subset])

        #JT: uses a method from MarginalsAnalyzerBase class, comments to that method can be found there

        self._marginals_averaged[subset] = self.get_averaged_marginal_for_subset(subset)

    #JT: We use this internal method to perform POVMs computation

    def _compute_POVM_from_marginals(self,
                                     marginals_dictionary: Dict[str, np.ndarray],
                                     qubit_indices: Union[str, List[int]], # why do we have two ways of specifying subsets
                                     estimation_method='pls'):
        # print('printing')
        # print(marginals_dictionary)
        # print(qubit_indices)
        # raise KeyboardInterrupt
        number_of_qubits = len(qubit_indices)
        dimension = int(2 ** number_of_qubits)

        #TODO FBM: fix this
        keys_list = sorted(list(marginals_dictionary.keys()))


        from qrem.functions_qrem import ancillary_functions as anf
        # keys_list = anf.register_names_qubits(qubit_indices=qubit_indices,
        #                                       )

        # print(marginals_dictionary)
        #First index - probe ket
        #Second index - outcome label
        frequencies_array = np.zeros((len(keys_list), dimension), dtype=complex)

        #TODO FBM: debug
        # frequencies_array = np.full((len(keys_list), dimension), 10**(-6), dtype=complex)

        # print(keys_list)

        from qrem.functions_qrem import ancillary_functions as anf

        probe_kets = []

        #JT: A loop over all keys of marginals dictionary

        for index_state in range(len(keys_list)):
            key_now = keys_list[index_state]
            if estimation_method.lower() in ['maximal_likelihood', 'ml']:
                probe_kets.append(self._get_tensored_ket(key_now))

            #JT: for PLS here a PLS estimator is computed

            elif estimation_method.lower() in ['least_squares', 'ls', 'projected_least_squares', 'pls']:
                #TODO FIX PLS FOR SIMULATIONS
                probe_kets.append(self._get_tensored_ket_LS(key_now))



            try:
                frequencies_array[index_state, :] = marginals_dictionary[key_now][:]


            except(IndexError):
                frequencies_array[index_state, :] = marginals_dictionary[key_now][:, 0]

            #JT: This is a different error than that above, solution appers to be the same
            except(ValueError):
                frequencies_array[index_state, :] = marginals_dictionary[key_now][:, 0]
            except(KeyError):
                pass

        # qprint_array(frequencies_array)

        #JT the classes below are used to get physical quantities out of POVM estimators

        setup_QDT = QDTCalibrationSetup(qubits_number=number_of_qubits,
                                        probe_kets=probe_kets,
                                        frequencies_array=frequencies_array,
                                        states_labels=keys_list)

        fitter_QDT = QDTFitter()


        #JT: Here physical POVM is reconstructed

        POVM_now = fitter_QDT.get_POVM_estimator(calibration_setup=setup_QDT,
                                                 method=estimation_method)

        # if isinstance(qubit_indices, list):
        #     qubit_indices = anf.get_qubits_keystring(qubit_indices)

        self._POVM_dictionary[qubit_indices] = POVM_now

    def _compute_subset_POVM(self,
                             subset,
                             estimation_method='pls'):
        # subset_key = 'q' + 'q'.join([str(s) for s in subset])

        #JT: the if below is redundant (at least for as we use it, as the same codition is checked in compute_subsets_POVMs_averaged)

        if subset not in self._marginals_averaged.keys():
            self._compute_averaged_marginals_subset(subset)

        averaged_marginals = self._marginals_averaged[subset]
        self._compute_POVM_from_marginals(averaged_marginals,
                                          qubit_indices=subset,
                                          estimation_method=estimation_method)

#JT This method is used to compute POVMs once marginals are computed
    def compute_subsets_POVMs_averaged(self,
                                       subsets_of_qubits: List[List[int]],
                                       estimation_method:str='pls',
                                       show_progress_bar: Optional[bool] = True) -> None:

        #Here number of subsets is established
        subsets_range = range(len(subsets_of_qubits))

        for subset in tqdm(subsets_of_qubits, disable = not show_progress_bar):
            #JT if subset is missing in averadged marginals dictionary then computation its performed

            if subset not in self._marginals_averaged.keys():
                self._compute_averaged_marginals_subset(subset)

        #JT: onec averadged marginals dictionary is ready computation of POVMs is done

        for subset_index in subsets_range:
            self._compute_subset_POVM(subsets_of_qubits[subset_index],
                                      estimation_method=estimation_method)

#JT: This method is used to get noise matrices from reconstructed POVMs, used e.g. after clustering to get noise matrices for computed clusters
# subsets list - list of subsets for which computation should be performed

    def compute_noise_matrices_from_POVMs(self,
                                       subsets_of_qubits: List[List[int]],
                                       show_progress_bar: Optional[bool] = True) -> None:

        for subset in tqdm(subsets_of_qubits,
                           disable = not show_progress_bar):

            #JT: computation of stochastic noise from a quantum POVM, unig a function from povmtools

            reduced_noise_matrix_now = povmtools.get_stochastic_map_from_povm(povm=self.POVM_dictionary[subset])

            #if subset in self.noise_matrices_dictionary[subset]:

            #JT: if key corresponding to a particular subset is present in noise_matrix_dictionary then internal dictionary is updated - (nested structure of disctionaries)
            # the key is subset and 'averaged'

            if subset in self.noise_matrices_dictionary.keys():
                self.noise_matrices_dictionary[subset]['averaged'] = reduced_noise_matrix_now

            #JT: If there is no entry in the dictionary, it is created with value corresponding to another dictionary

            else:
                self.noise_matrices_dictionary[subset]= {'averaged':reduced_noise_matrix_now}




import numpy as np
from qrem.common import io
from qrem.common import probability

from qrem.cn.simtools.auxiliary_merging_functions import divide_qubits_in_the_clusters as dq

import qrem.cn.simulation 

#Script for custom noise model creation, a loop goes through a list of specifications [cluster_size,number_of_clusters] and saves the models in files for future access
directory = '/home/kasiakm/Documents/QREM_DATA/Simulations100q/'
number_of_qubits = 100
#specifications = [[[4,25]],[[4,20],[2,10]],[[4,15],[3,10],[2,5]],[[3,30],[2,5]],[[4,1],[3,32]],[[4,10],[2,30]],[[3,33],[1,1]],[[3,24],[2,14]],[[2,50]],[[4,20],[3,6],[1,2]]]
specifications = [[[10,10]],[[5,20]]]


for s in specifications:
    count = 0
    for t in s:
        count+=t[0]*t[1]
    if(count!=number_of_qubits):
        print("WARNING!!! qubit number mismatch")


first_label_number = 0

for i in range(len(specifications)):
    model = qrem.cn.simulation.create_random_noise_model(number_of_qubits = number_of_qubits,clusters_specification = specifications[i])
    model_dict = model.get_dict_format()
    file_name = str(number_of_qubits)+'qmodel'+str(i+first_label_number)
    io.save(model_dict,
            directory=directory,
            custom_filename=file_name,
            overwrite=False)
    print(f"{i} done")






"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

from typing import Optional, Dict, Tuple, List
import abc
import numpy as np
from qrem.functions_qrem import functions_data_analysis as fda
from qrem.noise_characterization.base_classes.marginals_analyzer_base import MarginalsAnalyzerBase # used only to compute marginals
from qrem.noise_mitigation.base_classes.EnergyEstimatorInterface import EnergyEstimatorInterface
from tqdm import tqdm


class EnergyEstimatorBase(EnergyEstimatorInterface):
    """
    This is base class for estimating corrected energy of some Hamiltonian.

    In this class and its children, we use the following convention for:

     1. Generic experimental results:
    :param results_dictionary: Nested dictionary with following structure:

    results_dictionary[label_of_experiment][bitstring_outcome] = number_of_occurrences

    where:
        -label_of_experiment is arbitrary label for particular experiment,
        -bitstring_outcome is label for measurement outcome,
        -number_of_occurrences is number of times that bitstring_outcome was recorded

        Hence top-level key labels particular experiment
        (one can think about quantum circuit implementation)
        and its value is another dictionary with results of given experiment in the form
        of dictionary of measurement outcomes


    2. hamiltonian_dictionaries:
        :param hamiltonian_dictionaries: Nested dictionary with the following structure:

        hamiltonian_dictionaries[label_of_experiment][label_of_subset] = weight in decomposition
                                                                    of the Hamiltonian into Z-Paulis

        where:
            -label_of_experiment is the same as in results_dictionary and it labels results from which
            marginal distributions were calculated
            -label_of_subset is a label for qubits subset for which marginals_dictionary were calculated.
            We use convention that such label if of the form "q5q8q12..." etc., hence it is bitstring of
            qubits labels starting from "q".
            -marginal_probability_vector marginal distribution stored as vector

    """

    def __init__(self,
                 results_dictionary: Dict[str, Dict[str, int]],
                 hamiltonian_dictionaries: Dict[str, Dict[Tuple[int], float]],
                 energies_dictionary_raw: Optional[Dict[str, float]] = None
                 ) -> None:

        """
        :param results_dictionary: see class description
        :param hamiltonian_dictionaries: see class description
        """

        # set initial values of class properties
        self._results_dictionary = results_dictionary
        self._hamiltonian_dictionaries = hamiltonian_dictionaries

        if energies_dictionary_raw is None:
            energies_dictionary_raw = {}

        self._energies_dictionary_raw = energies_dictionary_raw

        self._energies_dictionary_corrected = {}


        results_of_exemplary_experiment = list(results_dictionary.values())[0]
        self._number_of_qubits = len(list(results_of_exemplary_experiment.keys())[0])

    @property
    def results_dictionary(self) -> Dict[str, Dict[str, int]]:

        return self._results_dictionary

    @results_dictionary.setter
    def results_dictionary(self, results_dictionary: Dict[str, Dict[str, int]]) -> None:
        self._results_dictionary = results_dictionary

    @property
    def hamiltonian_dictionaries(self) -> Dict[str, Dict[Tuple[int], float]]:
        return self._hamiltonian_dictionaries

    @hamiltonian_dictionaries.setter
    def hamiltonian_dictionaries(self,
                                 hamiltonian_dictionaries: Dict[
                                     str, Dict[Tuple[int], np.ndarray]]) -> None:
        self._hamiltonian_dictionaries = hamiltonian_dictionaries

    @property
    def energies_dictionary_raw(self) -> Dict[str, float]:

        return self._energies_dictionary_raw

    @property
    def energies_dictionary_corrected(self) -> Dict[str, float]:
        return self._energies_dictionary_raw

    def estimate_energy(self,
                        experiment_key: str,
                        method: Optional[str] = 'direct') -> float:

        experimental_results = self._results_dictionary[experiment_key]
        hamiltonian_weights = self._hamiltonian_dictionaries[experiment_key]

        if method.upper() == 'DIRECT':
            estimated_energy = fda.estimate_energy_from_counts_dictionary(
                counts_dictionary=experimental_results,
                weights_dictionary=hamiltonian_weights)

        elif method.upper() == 'MARGINALS':       # MOcomm loopy usage of marginal analyzer base
            marginals_analyzer = MarginalsAnalyzerBase(results_dictionary=self._results_dictionary,
                                                       bitstrings_right_to_left=False)
            estimated_marginals = marginals_analyzer.get_marginals(experiment_key=experiment_key,
                                                                   subsets_list=list(
                                                                       hamiltonian_weights.keys()))

            estimated_energy = fda.estimate_energy_from_marginals(
                weights_dictionary=hamiltonian_weights,
                marginals_dictionary=estimated_marginals)
        else:
            raise ValueError(f"Method name '{method}' unrecognized.\n "
                             f"Please choose either 'direct' or 'marginals' method.")

        self._energies_dictionary_raw[experiment_key] = estimated_energy

        return estimated_energy

    def estimate_multiple_energies(self,
                                   experiment_keys: List[str],
                                   method: Optional[str] = 'direct'):

        for experiment_label in tqdm(experiment_keys):
            self.estimate_energy(experiment_key=experiment_label,
                                 method=method)
        return self._energies_dictionary_raw




from collections import defaultdict
import copy
from typing import Tuple, List
from tqdm import tqdm

import numpy as np

from qrem.functions_qrem import ancillary_functions as anf, quantum_ancillary_functions as quanf
from qrem.functions_qrem import functions_data_analysis as fda
from qrem.functions_qrem import functions_probabilities as fun_prob
from qrem.noise_model_generation.CTMP import functions_CTMP as ctmp_fun
# from qrem.noise_simulation.CTMP.functions_CTMP_conditional import *

from qrem.common import convert, math
# from qrem.common.printer import qprint_array




#MOcomm - poorely documented, needs to be understood







def get_local_stochastic_matrix_from_generator(local_generator):
    return local_generator + np.eye(local_generator.shape[0])


def get_local_stochastic_matrix(operator_type):
    return get_local_stochastic_matrix_from_generator(
        ctmp_fun.return_generator_matrix(operator_type=operator_type))


def get_local_stochastic_matrices_from_rates(rates_dictionary):
    noise_matrices_dictionary = {qubits_tuple: {} for qubits_tuple in rates_dictionary.keys()}

    for qubits_tuple in rates_dictionary.keys():
        for operator_type in rates_dictionary[qubits_tuple].keys():
            noise_matrices_dictionary[qubits_tuple][operator_type] = get_local_stochastic_matrix(
                operator_type=operator_type)

    return noise_matrices_dictionary


def get_probability_distribution_over_error_rates(rates_dictionary,
                                                  normalization=None):
    if normalization is None:
        normalization = ctmp_fun.get_CTMP_error_rates_sum(rates_dictionary=rates_dictionary)

    probability_distribution = {qubits_tuple: {} for qubits_tuple in rates_dictionary.keys()}

    for qubits_tuple in rates_dictionary.keys():
        for operator_type in rates_dictionary[qubits_tuple].keys():
            probability_distribution[qubits_tuple][operator_type] = rates_dictionary[qubits_tuple][
                                                                        operator_type] / normalization

    return probability_distribution


def __sample_from_local_matrix_special(local_input_state: str,
                                       local_matrix: np.ndarray,
                                       ):
    probability_distribution = local_matrix[:, int(local_input_state, 2)]

    return convert.integer_to_bitstring(math.get_sample_from_multinomial_distribution(
        probability_distribution=probability_distribution), len(local_input_state))


def __flatten_probability_over_matrices(probability_over_matrices):
    probability_over_matrices_flattened = []

    for subset in probability_over_matrices.keys():
        for matrix_type in probability_over_matrices[subset].keys():
            probability_over_matrices_flattened.append(probability_over_matrices[subset][matrix_type])

    return probability_over_matrices_flattened


def __flatten_matrices_dictionary(local_noise_matrices_dictionary):
    local_noise_matrices_list = []

    for subset in local_noise_matrices_dictionary.keys():
        for matrix_type in local_noise_matrices_dictionary[subset].keys():
            local_noise_matrices_list.append(
                (subset, local_noise_matrices_dictionary[subset][matrix_type]))

    return local_noise_matrices_list


def _simple_sampling_one_step(global_input_state,
                              probability_over_matrices_list,
                              local_noise_matrices_list: List[Tuple[Tuple[int], np.ndarray]]):
    matrix_index = math.get_sample_from_multinomial_distribution(
        probability_distribution=probability_over_matrices_list)
    chosen_stuff = local_noise_matrices_list[matrix_index]

    chosen_subset, local_noise_matrix = chosen_stuff[0], chosen_stuff[1]
    local_input_state = ''.join([global_input_state[qi] for qi in chosen_subset])

    local_output_state = __sample_from_local_matrix_special(local_matrix=local_noise_matrix,
                                                            local_input_state=local_input_state)

    # print(local_noise_matrix)
    global_output_state = list(copy.deepcopy(global_input_state))

    for qi in range(len(chosen_subset)):
        global_output_state[chosen_subset[qi]] = local_output_state[qi]

    return ''.join(global_output_state)


def sample_from_power_of_B_matrix_simple(global_input_state,
                                         matrix_power,
                                         probability_over_matrices_list,
                                         local_noise_matrices_list
                                         ):
    output_state_now = copy.deepcopy(global_input_state)

    for power_index in range(matrix_power):
        output_state_now = _simple_sampling_one_step(global_input_state=output_state_now,
                                                     probability_over_matrices_list=probability_over_matrices_list,
                                                     local_noise_matrices_list=local_noise_matrices_list)

    return output_state_now

def _get_inputs_to_sampling_algorithm(rates_dictionary):
    noise_strength_CTMP = ctmp_fun.get_CTMP_error_rates_sum(rates_dictionary=rates_dictionary)
    local_noise_matrices = get_local_stochastic_matrices_from_rates(rates_dictionary=rates_dictionary)
    probability_over_matrices = get_probability_distribution_over_error_rates(
        rates_dictionary=rates_dictionary,
        normalization=noise_strength_CTMP)

    local_noise_matrices_flat = __flatten_matrices_dictionary(
        local_noise_matrices_dictionary=local_noise_matrices)
    probability_over_matrices_flat = __flatten_probability_over_matrices(
        probability_over_matrices=probability_over_matrices)

    return local_noise_matrices_flat, probability_over_matrices_flat


def sample_from_CTMP_noise_model_simple_method(global_input_state,
                                               rates_dictionary,
                                               number_of_samples):
    """

    Please note that this method requires setting value of CTMP normalization (gamma parameter from Ref.[#TODO FBM: add ref])
    to sum of error rates, which might be prohibitively big.

    :param global_input_state:
    :type global_input_state:
    :param rates_dictionary:
    :type rates_dictionary:
    :param number_of_samples:
    :type number_of_samples:
    :return:
    :rtype:

    """
    noise_strength_CTMP = ctmp_fun.get_CTMP_error_rates_sum(rates_dictionary=rates_dictionary)
    local_noise_matrices_flat, probability_over_matrices_flat = _get_inputs_to_sampling_algorithm(rates_dictionary)

    counts_dictionary = defaultdict(float)

    for _ in range(number_of_samples):
        random_power = np.random.poisson(noise_strength_CTMP,
                                         size=1)[0]

        output_state_now = sample_from_power_of_B_matrix_simple(global_input_state=global_input_state,
                                                                matrix_power=random_power,
                                                                probability_over_matrices_list=probability_over_matrices_flat,
                                                                local_noise_matrices_list=local_noise_matrices_flat)
        counts_dictionary[output_state_now] += 1

    return counts_dictionary



def estimate_corrected_energy_CTMP_simple(input_state,
                                          rates_dictionary,
                                          number_of_samples_ctmp,
                                          weights_dictionary_tuples):
    noisy_counts_dictionary_sampled_ctmp = sample_from_CTMP_noise_model_simple_method(
        global_input_state=input_state,
        number_of_samples=number_of_samples_ctmp,
        rates_dictionary=rates_dictionary
    )

    energy_modeled_ctmp_sampled = 0
    for output_state, number_of_times in noisy_counts_dictionary_sampled_ctmp.items():
        energy_modeled_ctmp_sampled += number_of_times * quanf.get_energy_from_bitstring_diagonal(
            bitstring=output_state,
            weights_dict=weights_dictionary_tuples)
    energy_modeled_ctmp_sampled /= number_of_samples_ctmp

    return energy_modeled_ctmp_sampled




# np.random.poisson(lambda_par)

#
# number_of_qubits = 8
#
# all_pairs = [(i, j) for i in range(number_of_qubits) for j in
#              range(i + 1, number_of_qubits)]
# single_qubits = [(qi,) for qi in range(number_of_qubits)]
# # all_pairs = all_pairs+[(i,) for i in range(number_of_qubits)]
# rates_2q = {subset: {x: np.random.uniform(0.01, 0.05) for x in ['00', '01', '10', '11']} for subset in
#             all_pairs}
# rates_1q = {subset: {x: np.random.uniform(0.01, 0.05) for x in ['0', '1']} for subset in single_qubits}
#
# rates_dictionary = {**rates_1q, **rates_2q}
# # print(rates_dictionary)
#
# # normalization_rates = ctmp_fun.get_CTMP_error_rates_sum(rates_dictionary=rates_dictionary)
#
# global_generator_matrix = ctmp_fun.get_global_generator_from_rates(rates_dictionary=rates_dictionary,
#                                                           number_of_qubits=number_of_qubits)
#
# # qprint_array(global_generator_matrix)
# global_noise_matrix = sc.linalg.expm(global_generator_matrix)
#
# classical_register = anf.get_classical_register_bitstrings(range(number_of_qubits))
#
# input_state = classical_register[-1]
#
# noisy_distro_exact = global_noise_matrix[:, int(input_state, 2)]
#
# number_of_samples = 10 ** 5
#
# # number_of_samples = int(np.ceil(number_of_samples_init*np.exp(normalization_rates*2)))
#
# # generators_dictionary = ctmp_fun.construct_model_from_rates(rates_dictionary=rates_dictionary)
#
# noisy_counts_dictionary_sampled = sample_from_CTMP_noise_model_simple_method(global_input_state=input_state,
#                                                                             rates_dictionary=rates_dictionary,
#                                                                             number_of_samples=number_of_samples)
# noisy_distro_sampled = fda.convert_counts_dictionary_to_probability_distribution(noisy_counts_dictionary_sampled)
#
# # from povms_qi import povmtools
# # povmtools.prob
#
# # print(noisy_distro_exact)
# # print(noisy_distro_sampled)
# from qrem.functions_qrem import povmtools
# print(povmtools.calculate_total_variation_distance(noisy_distro_exact,noisy_distro_sampled))
# # print(noisy_distro_sampled)



import os, shutil
from datetime import datetime
from typing import List, Union, Optional,Tuple
from pathlib import Path
import pickle

import numpy as np
from braket.aws import AwsDevice
from braket.circuits import Circuit

# qrem imports
from qrem.common.constants import EXPERIMENT_TYPE_SYMBLOS
from qrem.common.printer import qprint, errprint, warprint
from qrem.common.utils import get_full_object_size
from qrem.types import CircuitCollection
from braket.aws import AwsQuantumJob


def add_ground_states(groundstates_file = ""):
    pass

#-----------------------
# PART 1: Get backend properties (most importantly - good qubits)
#-----------------------
def get_device(device_full_name = "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-2",
               verbose_log = False):
    
    device = AwsDevice(device_full_name)
    qubit_count = device.properties.paradigm.qubitCount
    device_cost = device.properties.service.deviceCost
    execution_windows = device.properties.service.executionWindows
    connectivity_graph = device.properties.paradigm.connectivity
    calibration = device.properties.provider.specs

    if verbose_log:
        print('The maximum number of qubits supported by this device is', qubit_count)
        print('Execution windows', execution_windows)
        print('The price of running tasks on this device:', device_cost)

    metadata = {
        "device_name": device_full_name,
        "qubit_count": qubit_count,
        "execution_windows": execution_windows,
        "connectivity_graph": connectivity_graph,
        "calibration_data": calibration,
        "device_cost": device_cost,
    }
    return device, metadata

def get_valid_qubits_properties(  device: AwsDevice, 
                            threshold: Optional[float] = 0,
                            verbose_log = False,
                            simultaneous_randomize_benchmark: bool=True ):

    one_qubit_properties = device.properties.standardized.oneQubitProperties
    quantum_register_size = device.properties.paradigm.qubitCount

    # - quantum register size
    test_quantum_reg_size = len(one_qubit_properties)
    if test_quantum_reg_size != quantum_register_size:
        print(f"ERROR: test_quantum_reg_size <{test_quantum_reg_size}> does not match oneQubitProperties schema size<{quantum_register_size}>. Check braket code for changes.")

    one_qubit_fidelity_dict = {}
    qbits_above_error_treshold = [] # randomized benchmark
    qbits_above_error_treshold_simultaneous = [] #simultanous randomized benchmark
    all_qubits = [int(qubit_key) for qubit_key in one_qubit_properties]
    # [2] get qbits above treshold
    # TODO change to get exactly qubit_key list from one_qubit_properties
    
    if(threshold!= None and threshold > 0):
        # [2.1] for each qubit in backend get properties from one_qubit_properties:
        # https://amazon-braket-schemas-python.readthedocs.io/en/latest/_apidoc/braket.device_schema.standardized_gate_model_qpu_device_properties_v1.html
        for qubit_key in one_qubit_properties:
            #all_qubits.append(qubit_key)
            one_qubit_fidelity_dict[qubit_key] = {}
            one_qubit_fidelity_dict[qubit_key]['rb'] = one_qubit_properties[qubit_key].oneQubitFidelity[0].fidelity
            one_qubit_fidelity_dict[qubit_key]['simultaneous_rb'] = one_qubit_properties[qubit_key].oneQubitFidelity[1].fidelity

            if 1 - one_qubit_fidelity_dict[qubit_key]['rb'] > threshold:
                qbits_above_error_treshold.append(int(qubit_key))
            if 1 - one_qubit_fidelity_dict[qubit_key]['simultaneous_rb'] > threshold:
                qbits_above_error_treshold_simultaneous.append(int(qubit_key))
    else:
        
        qbits_above_error_treshold = []
        qbits_above_error_treshold_simultaneous = []
   
    if simultaneous_randomize_benchmark:
        all_bad_qubits = qbits_above_error_treshold_simultaneous;
    else:
        all_bad_qubits = qbits_above_error_treshold;

    good_qubits_indices =  [q for q in all_qubits if q not in all_bad_qubits]
    # [3.1] Renmove double entries and sort indices to ensure, that results can easily be interpreted. (NOT NECESSARY HERE MOST LIKELY)
    good_qubits_indices = list(set(good_qubits_indices)) 
    # [3.2] Important - indicies should be always sorted
    good_qubits_indices = sorted(good_qubits_indices) 
    number_of_good_qubits = len(good_qubits_indices)

    if(verbose_log):
        print(f"All {number_of_good_qubits} good qubit indices: {good_qubits_indices}")

    return {"good_qubits_indices": good_qubits_indices,
            "number_of_good_qubits": number_of_good_qubits,
            "original_quantum_register_size": quantum_register_size,
            "faulty_qubits": all_bad_qubits
            }


#-----------------------
# PART 2: Translate circuits to IBM / QISKIT FORMAT
#-----------------------
def _circuits_translator(eigenstate_index: int,
                            quantum_circuit: Circuit,
                            qubit_index: int):
    """
    Helper method that creates both DDOT and QDOT circuits in format native to BRAKET and applies to existing (empty) braket_circuit,
     by applying Pauli eigenstates (formerly _apply_pauli_eigenstate).
    """
    # _pauli_labels = ['z+', 'z-', 'x+', 'x-', 'y+', 'y-']

    # Z+
    if eigenstate_index == 0:
        quantum_circuit = quantum_circuit.rz(qubit_index, 0)

    # Z-
    elif eigenstate_index == 1:
        # quantum_circuit = quantum_circuit.rx(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rz(qubit_index, -0.9199372448290238)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rx(qubit_index, - np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, 2.2216554087607694)

    # X+
    elif eigenstate_index == 2:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)

    # X-
    elif eigenstate_index == 3:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, -np.pi / 2)

    # Y+
    elif eigenstate_index == 4:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)

    # Y-
    elif eigenstate_index == 5:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)

    else:
        raise ValueError(f"Incorrect eigenstate index: '{eigenstate_index}'!")

    return quantum_circuit

def translate_circuits_to_braket_format(circuit_collection:CircuitCollection,
                                        valid_qubit_indices:list) -> List[Circuit]:
    
    # TODO (PP) sort out repeating circuits already in CircuitCollection
    # prerequisite: no repeating circuits in the accepted collection
    braket_circuits = []
    
    for id in range(0,len(circuit_collection.circuits)):
        #[1] Make a label string from circuit labels - 
        #TODO (PP) won't work for machines with 1 million qubits, also now not used in AWS case
        base_label = np.array2string(circuit_collection.circuits[id], precision=0, separator='', max_line_width = 1000000)[1:-1]
        label = circuit_collection.experiment_name + "___" + base_label + "___" +str(id)

        #[2]  Make an empty circuit
        braket_circuit = Circuit()
        qrem_circuit=circuit_collection.circuits[id]

        for qubit_index in range(qrem_circuit.size):
            qubit_id = valid_qubit_indices[qubit_index]
            eigenstate_index = qrem_circuit[qubit_index]
            braket_circuit = _circuits_translator(eigenstate_index=eigenstate_index, 
                                                  quantum_circuit=braket_circuit,
                                                  qubit_index=qubit_id)

        braket_circuit_verbatim = Circuit().add_verbatim_box(braket_circuit)
        braket_circuits.append(braket_circuit)
    return braket_circuits



#-----------------------
# PART 3: Prepare braket circuits and all metadata to run adn output to job_dir
#-----------------------
def prepare_cricuits(braket_circuits:List[Circuit],
                     circuit_collection:CircuitCollection,
                     good_qubits_indices,
                     number_of_repetitions:int = 1024,
                     experiment_name:str = "",
                     metadata:dict = {},
                     job_tags:dict = {"QREM_JOB",},
                     job_dir:str = "",
                     pickle_submission = False,
                     number_of_task_retries:int = 3,
                     overwrite_output = False,
                     verbose_log:bool = False):
    
    # [0] default naming that always should be the same
    if(pickle_submission):
        circuits_file = 'braket_circuits_list.pkl'
    else:
        circuits_file = None
    qrem_circuits_file = 'qrem_circuits_list.txt'
    good_qubits_indices_file = 'good_qubits_indices.txt'
    metadata_file = 'metadata.pkl'

  
    # [1] Check and prepare output folder
    job_path= Path(job_dir)
    if not job_path.is_dir():
        job_path.mkdir(parents=True,exist_ok=False)
    elif any(job_path.iterdir()) == True and overwrite_output:
        try:
            shutil.rmtree(str(job_path))
        except OSError:
            os.remove(str(job_path))
        job_path.mkdir(parents=True,exist_ok=False)
    elif  any(job_path.iterdir()) == True and not overwrite_output:
        print(f"ERROR: non empty directory would be overwritten {str(job_path)}. Cancelling")
        return False
    else:
        print(f"ERROR: usnpecified. Cancelling")
        return False
    
    
    # [1.1] Check and prepare output fules
    circuits_file_path = job_path.joinpath(circuits_file)
    qrem_circuits_file_path = job_path.joinpath(qrem_circuits_file)
    good_qubits_indices_file_path = job_path.joinpath(good_qubits_indices_file)
    metadata_file_path = job_path.joinpath(metadata_file)

    # [2] write the circuit symbols to pickle file:
    if(pickle_submission):
        with open(circuits_file_path, 'wb') as fp:  # open a text file
            pickle.dump(braket_circuits, fp) # serialize the list

    # write the circuit symbols to txt file:
    with open(qrem_circuits_file_path, 'w') as fp:
        for circuit in circuit_collection.circuits:
            for label in circuit:
                fp.write("%s" % str(label))
            # write each item on a new line
            fp.write("\n")

    # [3] write the used qubits to txt file:
    with open(good_qubits_indices_file_path, 'w') as fp:
        for qubit_index in good_qubits_indices:
            fp.write("%s\n" % qubit_index)
            


    # [4] Prepare and save the metadata to pickle file:
    now = datetime.now()
    metadata["date"] = f"{now.year}-{now.month}-{now.day}"
    metadata["number_of_repetitions"] = number_of_repetitions
    metadata["max_task_retries"] = number_of_task_retries

    # [4.1] Prepare and save tthe tags to metadata:
    job_tags_dict = {tag:tag for tag in job_tags.items()}
    job_tags_dict["experiment"] = experiment_name
    job_tags_dict["date"] = metadata["date"]
    metadata["tags"] = job_tags_dict


    with open(metadata_file_path, 'wb') as fp:  # open a text file
        pickle.dump(metadata, fp) # serialize the metadata
        
    return True

#-----------------------
# PART 4: Execute circuits
#-----------------------
def execute_circuits(
    device_name:str="arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-2",
    pickle_submission= False,
    job_dir:Path="job_dir",
    verbose_log = True):
    """
    Packs all data and script runner and executes on AWS machine
    """

    #[0] Define all path based on the job dir
    current_script_dir = Path( __file__ ).parent.absolute()

    if(pickle_submission):
        path_to_circuits = job_dir.joinpath('braket_circuits_list.pkl')
    qrem_path_to_circuits = job_dir.joinpath('qrem_circuits_list.txt')
    path_to_module = current_script_dir.joinpath("awsutils").joinpath("aws_braket_runner.py")
    runner_function = "aws_braket_runner:run"

    abspath_to_circuits = str(path_to_circuits.resolve())
    abspath_to_qrem_circuits = str(qrem_path_to_circuits.resolve())
    abspath_to_module = str(path_to_module.resolve())

    path_to_good_qubits_indices = job_dir.joinpath('good_qubits_indices.txt')
    abspath_to_good_qubits_indices = str(path_to_good_qubits_indices.resolve())

    path_to_metadata = job_dir.joinpath('metadata.pkl')
    abspath_to_metadata = str(path_to_metadata.resolve())


    #[1] Define input data
    input_data={'qrem_circuits_list': abspath_to_qrem_circuits,
                'good_qubits_indices': abspath_to_good_qubits_indices,
                'metadata': abspath_to_metadata}
    
    if(pickle_submission):
        input_data['braket_circuits_list'] = abspath_to_circuits
    
    #[2] Create a job
    job = AwsQuantumJob.create(
        device_name,
        source_module=abspath_to_module,
        input_data=input_data,
        entry_point=runner_function,
        wait_until_complete=False
    )
    if(verbose_log):
        print(job)





#MOcomm - poorely documented, needs to be understood
import numpy as np
from collections import defaultdict
import random
from typing import Dict, Tuple, List
from tqdm import tqdm
import copy

def generate_binary_strings(n, l):
    if n == 0:
        return l
    else:
        if len(l) == 0:
            return generate_binary_strings(n - 1, ["0", "1"])
        else:
            return generate_binary_strings(n - 1, [i + "0" for i in l] + [i + "1" for i in l])

def kronecker_delta(v, w):
    if v==w:
        return 1
    else:
        return 0


def get_cond_prob(x, y,
                  generators_dictionary:Dict[Tuple[int],
                                             np.ndarray],
                  normalization_quasiprobability, printing=False):
    if printing: print('look for ' + str(x) + '->' + str(y))
    val = 0
    number_of_qubits = len(y)
    all_qubits_indices = set(range(number_of_qubits))
    if printing: print('supp: ' + str(all_qubits_indices))
    val += kronecker_delta(y, x[:number_of_qubits])
    if val == 1:
        if printing: print('DIAGONAL')
    for local_qubits_tuple in generators_dictionary.keys():
        operator = generators_dictionary[local_qubits_tuple]
        if printing: print(operator)
        # op_support_ordered = [int(qubit) for qubit in op_support_ordered_str.split('q')[1:]]
        if printing: print(local_qubits_tuple)

        op_support_cap_support = set(local_qubits_tuple).intersection(set(all_qubits_indices))
        if not op_support_cap_support == set():
            support_cond = True
            if printing: print('oscs: ' + str(op_support_cap_support))
            support_minus_op_support = set(all_qubits_indices).difference(set(local_qubits_tuple))
            for qubit in support_minus_op_support:
                if y[qubit] != x[qubit]:
                    support_cond = False
                    if printing: print('sc failed for qubit' + str(qubit))
                    break
            if (support_cond):
                if printing: print('sc OK')
                op_support_minus_support = set(local_qubits_tuple).difference(set(all_qubits_indices))
                if printing: print('osms: ' + str(op_support_minus_support))
                y_supp_vals = dict()

                for qubit in op_support_cap_support:
                    y_supp_vals[qubit] = y[qubit]

                no_of_y_summed = len(op_support_minus_support)

                if (no_of_y_summed > 0):
                    y_summed_bitstrings = generate_binary_strings(no_of_y_summed, [])
                    for y_summed_bitstr in y_summed_bitstrings:
                        i = 0
                        if printing: print('checking bitstr' + str(y_summed_bitstr))
                        for qubit in op_support_minus_support:
                            y_supp_vals[qubit] = y_summed_bitstr[i]
                            i += 1
                        y_supp_vals_ordered = "".join("%s" % y_supp_vals[qubit] for qubit in local_qubits_tuple)
                        row = int(y_supp_vals_ordered, 2)
                        x_supp_vals_ordered = "".join("%s" % x[qubit] for qubit in local_qubits_tuple)
                        col = int(x_supp_vals_ordered, 2)
                        if printing: print('added val ' + str(operator[row][col] / normalization_quasiprobability))
                        val += operator[row][col] / normalization_quasiprobability
                else:
                    y_supp_vals_ordered = "".join("%s" % y[qubit] for qubit in local_qubits_tuple)
                    row = int(y_supp_vals_ordered, 2)
                    x_supp_vals_ordered = "".join("%s" % x[qubit] for qubit in local_qubits_tuple)
                    col = int(x_supp_vals_ordered, 2)
                    if printing: print('no added bitstr')
                    if printing: print('added val ' + str(operator[row][col] / normalization_quasiprobability))
                    val += operator[row][col] / normalization_quasiprobability

    return val

def sample_from_B_matrix(input_state, generators_dictionary, quasiprobability_normalization, printing=False):
    y = ''
    no_of_bits = len(input_state)
    for bit_no in range(no_of_bits):
        y_0 = y + '0'
        y_1 = y + '1'
        if printing: print('testing ' + y_0 + ' and ' + y_1)
        cp_0 = get_cond_prob(input_state, y_0, generators_dictionary, quasiprobability_normalization)
        cp_1 = get_cond_prob(input_state, y_1, generators_dictionary, quasiprobability_normalization)
        tot = cp_0 + cp_1
        p_0 = cp_0 / tot
        if printing: print('p_0 ' + str(p_0))
        val = random.uniform(0, 1)
        if val < p_0:
            y = y + '0'
            if printing: print('extending ' + y)
        else:
            y = y + '1'
            if printing: print('extending ' + y)
    return y


def sample_from_power_of_B_matrix(matrix_power,
                                  input_state,
                                  generators_dictionary,
                                  noise_strength_CTMP,
                                  printing=False):
    output_state = copy.deepcopy(input_state)
    for step in range(matrix_power):
        if printing: print('pre ' + output_state)
        output_state = sample_from_B_matrix(output_state, generators_dictionary, noise_strength_CTMP)
        if printing: print('post ' + output_state)
    return output_state

def sample_from_CTMP_noise_model_conditional_probabilities(input_state:str,
                                                           number_of_samples:int,
                                                           generators_dictionary:Dict[Tuple[int],
                                                              np.ndarray],
                                                           normalization_quasiprobability:float):
    samples_from_noise_model = defaultdict(float)
    for _ in range(number_of_samples):
        random_power = np.random.poisson(normalization_quasiprobability)
        samples_from_noise_model[sample_from_power_of_B_matrix(random_power,
                                                               input_state,
                                                               generators_dictionary,
                                                               normalization_quasiprobability)]+=1
    return samples_from_noise_model


"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com

REFERENCES:
[0] Filip B. Maciejewski, Zoltán Zimborás, Michał Oszmaniec,
"Mitigation of readout noise in near-term quantum devices
by classical post-processing based on detector tomography",
Quantum 4, 257 (2020)

[0.5] Filip B. Maciejewski, Flavio Baccari, Zoltán Zimborás, Michał Oszmaniec,
"Modeling and mitigation of cross-talk effects in readout noise
with applications to the Quantum Approximate Optimization Algorithm",
Quantum 5, 464 (2021).

"""

from typing import Optional, Dict, Union, List, Tuple

import numpy as np

from qrem.functions_qrem import povmtools, ancillary_functions as anf
from qrem.noise_model_generation.CN.GlobalNoiseMatrixCreator import \
    GlobalNoiseMatrixCreator
from qrem.noise_model_generation.CN.NoiseModelGenerator import NoiseModelGenerator

from qrem.common import utils
from qrem.common.printer import qprint

class CorrectionDataGenerator(NoiseModelGenerator):
    """
        Main class used to calculate data needed for noise-mitigation on marginals_dictionary, based on provided
        noise model.

        NOTE: Currently it handles properly only two-qubit marginals_dictionary (e.g., experiments involving
              estimation of 2-local Hamiltonians)

        The correction data consists of the following:
        - :param: 'correction_indices' - for each marginal of interest (e.g., 'q0q1'), specify label for
                                 marginal that needs to be corrected and then coarse-grained in order
                                 to perform noise-mitigation. For example, if q0 is strongly correlated
                                 with q5 (i.e., they are in the same cluster), it is advisable to first
                                 perform noise mitigation on marginal 'q0q1q5', and then coarse-grain
                                 it to obtain 'q0q1'.

                                 The format we use here is dictionary where KEY is label for marginal
                                 of interest, and VALUE is label for marginal that needs to be
                                 calculated first.
                                 So the entry for example above would look like:
                                 correction_indices['q0q1'] = 'q0q1q5'


        - :param: 'noise_matrices' - the noise matrices representing effective noise matrix acting on marginals_dictionary
                            specified by values of correction_indices dictionary

                            This is dictionary where KEY is subset label and VALUE is noise matrix


        - :param: 'correction_matrices' - inverses of noise_matrices, convention used is the same
    """

    # TODO FBM: generalize for more than two-qubit subsets
    # TODO FBM: add mitigation errors

    def __init__(self,
                 results_dictionary_ddt: Dict[str, Dict[str, int]],
                 number_of_qubits: int,
                 bitstrings_right_to_left: Optional[bool] = False,
                 marginals_dictionary: Optional[Dict[str, Dict[Tuple[int], np.ndarray]]] = None,
                 clusters_list: Optional[List[Tuple[int]]] = None,
                 neighborhoods: Optional[Dict[Tuple[int], Tuple[int]]] = None,
                 noise_matrices_dictionary: Optional[
                     Dict[Tuple[int], Dict[Union[str, Tuple[int]], Dict[str, np.ndarray]]]] = None,
                 correction_matrices_dictionary: Optional[Dict[Tuple[int], np.ndarray]] = None
                 ) -> None:

        super().__init__(results_dictionary_ddt,
                         bitstrings_right_to_left,
                         marginals_dictionary,
                         noise_matrices_dictionary,
                         clusters_list,
                         neighborhoods
                         )
        self._number_of_qubits = number_of_qubits
        self._qubit_indices = range(number_of_qubits)

        if clusters_list is None:
            clusters_lists_dictionary, clusters_labels_dictionary = {}, {}

        #JT: clusters_lists_dictionary: a dictionary with keys corresponding to qubits and values to clusters 
        else:
            clusters_lists_dictionary = {(qi,): () for qi in self._qubit_indices}
            for cluster in self._clusters_list:
                for qi in cluster:
                    clusters_lists_dictionary[(qi,)] = cluster


            #JT: how this is different from clusters_lists dictionary
            clusters_labels_dictionary = {
                (qi,): clusters_lists_dictionary[(qi,)] for qi in
                self._qubit_indices}

        # self._noise_matrices = noise_matrices_dictionary
        self._correction_matrices = correction_matrices_dictionary

        if noise_matrices_dictionary is None:
            self._noise_matrices_dictionary = {}

        if correction_matrices_dictionary is None:
            self._correction_matrices = {}

        self._mitigation_errors = {}
        self._correction_indices = {}

        self._clusters_lists_dictionary = clusters_lists_dictionary
        self._clusters_labels_dictionary = clusters_labels_dictionary

    # @property
    # def correction_matrices(self)->Dict[Tuple[int],np.ndarray]:
    #

    def _set_clusters_dictionary(self) -> None:
        """
        Function that updates class' properties needed (later) to calculate correction data
        """

        clusters_lists_dictionary = {(qi,): () for qi in self._qubit_indices}
        for cluster in self._clusters_list:
            for qi in cluster:
                clusters_lists_dictionary[(qi,)] = cluster
        clusters_labels_dictionary = {
            (qi,): clusters_lists_dictionary[(qi,)] for qi in
            self._qubit_indices}
        self._clusters_lists_dictionary = clusters_lists_dictionary
        self._clusters_labels_dictionary = clusters_labels_dictionary

    def _construct_local_full_noise_model(self,
                                          qubit_i: int,
                                          qubit_j: int,
                                          intersection_i: Optional[List[int]] = None,
                                          intersection_j: Optional[List[int]] = None
                                          ) -> np.ndarray:
        """
        This is function to obtain local noise model for two clusters to which both qubits belong.
        It is slightly tedious, the main issue here is to properly sort qubits and format input data
        for GlobalNoiseMatrixCreator (different class)

        :param qubit_i: qubit that belongs to cluster i
        :param qubit_j: as above
        :param intersection_i: qubits that are neighbors of cluster i and belong to cluster j
        :param intersection_j: as above
        :return:
        """

        # get needed information from class' properties
        cluster_i = self._clusters_lists_dictionary[(qubit_i,)]
        cluster_j = self._clusters_lists_dictionary[(qubit_j,)]

        string_cluster_i = self._clusters_labels_dictionary[(qubit_i,)]
        string_cluster_j = self._clusters_labels_dictionary[(qubit_j,)]

        neighbors_i = self._neighborhoods[string_cluster_i]
        neighbors_j = self._neighborhoods[string_cluster_j]

        # if not provided, calculate it
        if intersection_i is None:
            intersection_i = utils.lists_intersection(neighbors_i, cluster_j)
        if intersection_j is None:
            intersection_j = utils.lists_intersection(neighbors_j, cluster_i)

        # Take cluster noise matrices depending on neighbors states
        matrices_cluster_i = self.get_noise_matrix_dependent(cluster_i,
                                                             neighbors_i)
        matrices_cluster_j = self.get_noise_matrix_dependent(cluster_j,
                                                             neighbors_j)

        # average over neighbours of first cluster
        # that do not include the members of second cluster
        averaged_matrices_cluster_i = self.average_noise_matrices_over_some_qubits(
            matrices_cluster_i,
            neighbors_i,
            intersection_i
        )
        # and vice versa
        averaged_matrices_cluster_j = self.average_noise_matrices_over_some_qubits(
            matrices_cluster_j,
            neighbors_j,
            intersection_j)

        """
        Here we will need to properly relabel qubits because noise matrix creator requires 
        numbering them from 0 to number of qubits.
        """
        # Label qubits from 0
        enumerated_map_reversed = utils.map_index_to_order(
            cluster_i + cluster_j)

        qubit_indices_enumerated = []
        for cluster_now in [cluster_i, cluster_j]:
            qubit_indices_enumerated.append(
                [enumerated_map_reversed[ci] for ci in cluster_now])

        cluster_key_for_construction_i = anf.get_qubits_keystring(
            qubit_indices_enumerated[0])
        cluster_key_for_construction_j = anf.get_qubits_keystring(
            qubit_indices_enumerated[1])

        # this is format of noise matrices dictionary accepted by GlobalNoiseMatrixCreator
        properly_formatted_lambdas = {cluster_key_for_construction_i: averaged_matrices_cluster_i,
                                      cluster_key_for_construction_j: averaged_matrices_cluster_j
                                      }

        # this is format for neighbors dictionary accepted by GlobalNoiseMatrixCreator
        neighbors_for_construction = {
            cluster_key_for_construction_i: [enumerated_map_reversed[ci] for ci in
                                             intersection_i],
            cluster_key_for_construction_j: [enumerated_map_reversed[cj] for cj in
                                             intersection_j]}

        # get instance of GlobalNoiseMatrixCreator
        big_lambda_creator_now = GlobalNoiseMatrixCreator(
            properly_formatted_lambdas,
            qubit_indices_enumerated,
            neighbors_for_construction)

        # get noise matrix
        local_noise_matrix = \
            big_lambda_creator_now.compute_global_noise_matrix()

        return local_noise_matrix

    @staticmethod
    def _sort_qubits_matrix(
            local_noise_matrix: np.ndarray,
            cluster_i: List[int],
            cluster_j: List[int]) -> np.ndarray:
        """
        :param local_noise_matrix: noise matrix acting on cluster_i and cluster_j
        :param cluster_i: list of qubits' indices
        :param cluster_j: list of qubits' indices
        :return: permuted local_noise_matrix
        """

        all_qubits_list = list(cluster_i + cluster_j)

        # This is our target
        sorted_qubits_dictionary = utils.enumerate_dict(sorted(all_qubits_list))
        # This is what we have
        qubits_dictionary = utils.enumerate_dict(all_qubits_list)

        # While what we have is not equal to target we sort qubits.
        # The following loop performs series of SWAPs to properly order qubits.
        while qubits_dictionary != sorted_qubits_dictionary:
            for index_qubit_here in range(len(all_qubits_list) - 1):
                if all_qubits_list[index_qubit_here + 1] < all_qubits_list[index_qubit_here]:
                    # if two qubits are not sorted in ascending order, we permute matrix
                    # this corresponds to exchanging place of two qubits in the Hilbert space
                    local_noise_matrix = povmtools.permute_matrix(local_noise_matrix,
                                                                  len(all_qubits_list),
                                                                  [index_qubit_here + 1,
                                                                   index_qubit_here + 2])
                    # update indices to keep track of already made swaps
                    all_qubits_list[index_qubit_here], all_qubits_list[
                        index_qubit_here + 1] = all_qubits_list[index_qubit_here + 1], all_qubits_list[
                        index_qubit_here]
            # update whole dictionary
            qubits_dictionary = utils.enumerate_dict(all_qubits_list)
        return local_noise_matrix

    def _compute_pair_noise_matrix(self,
                                   pair: Tuple[int]):

        qubit_i, qubit_j = pair[0], pair[1]

        cluster_i = self._clusters_lists_dictionary[(qubit_i,)]
        cluster_j = self._clusters_lists_dictionary[(qubit_j,)]

        if cluster_i == cluster_j:
            # Check if qubits are in the same cluster. If yes, we just take average
            # noise matrix on that cluster.
            averaged_matrix_clusters_i_j = self.get_noise_matrix_averaged(
                sorted(utils.lists_sum(cluster_i, cluster_j)))

        else:
            string_cluster_i = self._clusters_labels_dictionary[(qubit_i,)]
            string_cluster_j = self._clusters_labels_dictionary[(qubit_j,)]
            # If qubits are in different clusters, we have two options to consider below...
            intersection_i, intersection_j = \
                utils.lists_intersection(
                    self._neighborhoods[string_cluster_i],
                    cluster_j), \
                utils.lists_intersection(
                    self._neighborhoods[string_cluster_j],
                    cluster_i)

            if len(intersection_i) == 0 and len(intersection_j) == 0:
                # Check if clusters contain each others neighbors.
                # If not, the noise matrix is simply a
                # tensor product of cluster matrices
                averaged_matrix_cluster_i = self.get_noise_matrix_averaged(cluster_i)
                averaged_matrix_cluster_j = self.get_noise_matrix_averaged(cluster_j)

                # print(averaged_matrix_cluster_i,averaged_matrix_cluster_j)

                averaged_matrix_clusters_i_j = np.kron(averaged_matrix_cluster_i,
                                                       averaged_matrix_cluster_j)

            else:
                # Check if clusters are each others neighbors.
                # If yes, the noise matrix needs to be constructed using
                # cluster-neighborhoods noise model with treating
                # some members of clusters as neighbours
                averaged_matrix_clusters_i_j = self._construct_local_full_noise_model(
                    qubit_i=qubit_i,
                    qubit_j=qubit_j,
                    intersection_i=intersection_i,
                    intersection_j=intersection_j)

        local_noise_matrix = averaged_matrix_clusters_i_j
        # check whether qubits are properly sorted
        if cluster_i != cluster_j and cluster_i + cluster_j != sorted(cluster_i + cluster_j):
            # if qubits are not sorted, noise matrix needs to be permuted
            local_noise_matrix = self._sort_qubits_matrix(
                local_noise_matrix=averaged_matrix_clusters_i_j,
                cluster_i=cluster_i,
                cluster_j=cluster_j)

        return local_noise_matrix

    def _compute_pair_correction_data(self,
                                      pair: Tuple[int]) -> None:
        """
        For given pair of qubits, get correction data required
        to correct corresponding two-qubit marginal.
        NOTE: see class' description

        :param pair: list of qubit indices
        """

        qubit_i, qubit_j = pair[0], pair[1]

        cluster_i = self._clusters_lists_dictionary[(qubit_i,)]
        cluster_j = self._clusters_lists_dictionary[(qubit_j,)]

        whole_marginal = sorted(utils.lists_sum(cluster_i, cluster_j))
        marginal_indices_tuple = tuple(whole_marginal)

        # print(marginal_indices_tuple)

        # print(self._noise_matrices_dictionary.keys())
        if marginal_indices_tuple in self._noise_matrices_dictionary.keys():

            if 'averaged' in self._noise_matrices_dictionary[marginal_indices_tuple].keys():
                local_noise_matrix = self._noise_matrices_dictionary[marginal_indices_tuple]['averaged']
                # print('not computing noise matrix!')
            else:
                local_noise_matrix = self._compute_pair_noise_matrix(pair)

                self._noise_matrices_dictionary[marginal_indices_tuple]['averaged'] = local_noise_matrix
        else:
            #JT: A potential source of the problem
            local_noise_matrix = self._compute_pair_noise_matrix(pair)
            self._noise_matrices_dictionary[marginal_indices_tuple] = {'averaged': local_noise_matrix}

        # JT this is a quic try to fix problem with inverse, it's non-optimal and should be changed in the future, it will not work eg. in the case when nieghbours are considered
        # the overall matriox is constructed as a kronecker product of two inverses
        # firstly we calculate average noise matrices for the two qubits
        averaged_matrix_cluster_i = self.get_noise_matrix_averaged(cluster_i)

        #if the two qubits belong to the same cluster, the inverse is computed
        if cluster_i == cluster_j:
            try:
                correction_matrix = np.linalg.inv(averaged_matrix_cluster_i)

            except:
                qprint(
                    f"MATRIX i FOR SUBSET: {marginal_indices_tuple} is not invertible, computing pseudoinverse", '',
                    'red')
                correction_matrix = np.linalg.pinv(averaged_matrix_cluster_i)

        #if the qubits are in different clusters we need the two inverses
        else:
            averaged_matrix_cluster_j = self.get_noise_matrix_averaged(cluster_j)
            try:
                correction_matrix_cluster_i = np.linalg.inv(averaged_matrix_cluster_i)
            except:
                qprint(f"MATRIX i FOR SUBSET: {marginal_indices_tuple} is not invertible, computing pseudoinverse",'','red')
                correction_matrix_cluster_i = np.linalg.pinv(averaged_matrix_cluster_i)
            try:
                correction_matrix_cluster_j = np.linalg.inv(averaged_matrix_cluster_j)
            except:
                qprint(f"MATRIX j FOR SUBSET: {marginal_indices_tuple} is not invertible, computing pseudoinverse",
                           '', 'red')
                correction_matrix_cluster_j = np.linalg.pinv(averaged_matrix_cluster_j)
            correction_matrix = np.kron(correction_matrix_cluster_i, correction_matrix_cluster_j)
            #the order of qubits is checked, reordering is performed when needed
            if cluster_i + cluster_j != sorted(cluster_i+ cluster_j):
                correction_matrix = self._sort_qubits_matrix(local_noise_matrix=correction_matrix,cluster_i=cluster_i,cluster_j=cluster_j)
        #here the modification of inverse computation ends

        #JT uncimment this try except if the above does not work
        #try:
        #    correction_matrix = np.linalg.inv(local_noise_matrix)
        #except:
        #    qprint(f"MATRIX FOR SUBSET: {marginal_indices_tuple} is not invertible, computing pseudoinverse",'','red')
        #    qprint(f"DET: {np.linalg.det(local_noise_matrix)} ",'', 'red')
        #    correction_matrix = np.linalg.pinv(local_noise_matrix)

        #here modification ends


        # correction_matrix = np.linalg.inv(local_noise_matrix)

        #Perhaps do not do this
        self._correction_indices[(qubit_i,)] = marginal_indices_tuple
        self._correction_indices[(qubit_j,)] = marginal_indices_tuple
        self._correction_indices[(qubit_i, qubit_j)] = marginal_indices_tuple
        self._correction_matrices[marginal_indices_tuple] = correction_matrix

    def get_pairs_correction_data(self,
                                  pairs_list: List[Tuple[int]],
                                  show_progress_bar: Optional[bool] = False,
                                  reset_everything: Optional[bool] = False
                                  ) -> dict:
        """
        For pairs of qubits in the list, get correction data required
        to correct corresponding two-qubit marginals_dictionary.
        NOTE: see class' description for details

        :param pairs_list:
        :param show_progress_bar:
        :return: correction_data_dictionary
        """

        # TODO FBM: possibly change resetting
        self._set_clusters_dictionary()

        if reset_everything:
            # self._noise_matrices = {}
            self._correction_matrices = {}

        self._mitigation_errors = {}
        self._correction_indices = {}

        range_pairs = range(len(pairs_list))

        if show_progress_bar:
            from tqdm import tqdm
            range_pairs = tqdm(range_pairs)

        for pair_index in range_pairs:
            pair = pairs_list[pair_index]
            pair_key = tuple(pair)
            if pair_key not in self._correction_indices.keys():
                self._compute_pair_correction_data(pair)

        for qubit in range(self._number_of_qubits):
            # TODO FBM: perhaps modify this

            cluster_now = self._clusters_labels_dictionary[(qubit,)]
            self._correction_indices[(qubit,)] = cluster_now

            if len(cluster_now) == 1:
                if (qubit,) not in self._correction_matrices.keys():
                    # local_noise_matrix = self._compute_noise_matrix_averaged(subset=(qubit,))

                    local_noise_matrix = self.get_noise_matrix_averaged(subset=(qubit,))
                    # self.noise_matrices_dictionary[(qubit,)]['averaged'] = local_noise_matrix
                    try:
                        self._correction_matrices[(qubit,)] = np.linalg.inv(local_noise_matrix)
                    except:
                        qprint(f"MATRIX FOR SUBSET: {(qubit,)} is not invertible, computing pseudinverse",'','red')
                        self._correction_matrices[(qubit,)] = np.linalg.pinv(local_noise_matrix)


        return {'correction_matrices': self._correction_matrices,
                'noise_matrices': self.noise_matrices_dictionary,
                'correction_indices': self._correction_indices}
